import sys

from django.apps import apps

from lazy_ninja.registry import ModelRegistry, controller_for


def reset_registry(monkeypatch):
    monkeypatch.setattr(ModelRegistry, "_controllers", {})
    monkeypatch.setattr(ModelRegistry, "_discovered", True)


def test_register_and_get_controller(monkeypatch):
    reset_registry(monkeypatch)

    class DummyController:
        pass

    ModelRegistry.register_controller("TestModel", DummyController)
    assert ModelRegistry.get_controller("TestModel") is DummyController
    assert ModelRegistry.get_controller("unknown") is None


def test_controller_for_decorator_registers(monkeypatch):
    reset_registry(monkeypatch)

    @controller_for("Widget")
    class WidgetController:
        pass

    assert ModelRegistry.get_controller("Widget") is WidgetController


def test_discover_controllers_imports_modules(tmp_path, monkeypatch):
    reset_registry(monkeypatch)
    monkeypatch.setattr(ModelRegistry, "_discovered", False)

    package_dir = tmp_path / "sample_app"
    controllers_dir = package_dir / "controllers"
    controllers_dir.mkdir(parents=True)

    (package_dir / "__init__.py").write_text("")
    (controllers_dir / "__init__.py").write_text("")
    (controllers_dir / "widgets.py").write_text("registered = True\n")

    monkeypatch.syspath_prepend(str(tmp_path))

    class StubAppConfig:
        name = "sample_app"
        path = str(package_dir)

    monkeypatch.setattr(apps, "get_app_configs", lambda: [StubAppConfig])

    try:
        ModelRegistry.discover_controllers()
        assert ModelRegistry._discovered is True
        assert "sample_app.controllers.widgets" in sys.modules
    finally:
        sys.modules.pop("sample_app", None)
        sys.modules.pop("sample_app.controllers", None)
        sys.modules.pop("sample_app.controllers.widgets", None)



from typing import Any, Type, Optional, Callable
from asgiref.sync import sync_to_async

from ninja import Schema

from ..utils import serialize_model_instance, serialize_model_instance_async


class BaseResponseHandler:
    """Base class for response handlers."""
    
    def _apply_custom_response(self, custom_response: Callable, request: Any, instance: Any) -> Any:
        """Apply custom response if provided."""
        if custom_response:
            return custom_response(request, instance)
        return None


class SyncResponseHandler(BaseResponseHandler):
    """Handles response formatting for sync routes."""
    
    def handle_response(
        self, 
        instance: Any, 
        schema: Type[Schema], 
        custom_response: Optional[Callable] = None, 
        request: Any = None
    ) -> Any:
        """
        Handle the response formatting based on custom_response or schema validation.
        
        Args:
            instance: The instance to format
            schema: The schema to use for validation
            custom_response: Optional custom response handler
            request: Optional request object for custom response
            
        Returns:
            Formatted response
        """
        custom_result = self._apply_custom_response(custom_response, request, instance)
        if custom_result is not None:
            return custom_result
        
        serialized = serialize_model_instance(instance)
        return serialized


class AsyncResponseHandler(BaseResponseHandler):
    """Handles response formatting for async routes."""
    
    async def handle_response(
        self, 
        instance: Any, 
        schema: Type[Schema], 
        custom_response: Optional[Callable] = None, 
        request: Any = None
    ) -> Any:
        """
        Async version of handle_response that uses serialize_model_instance_async
        
        Args:
            instance: The instance to format
            schema: The schema to use for validation
            custom_response: Optional custom response handler
            request: Optional request object for custom response
            
        Returns:
            Formatted response
        """
        if custom_response:
            return await sync_to_async(custom_response)(request, instance)
        
        serialized = await serialize_model_instance_async(instance)
        return serialized


"""
Utility modules for lazy-ninja.
"""

# Core utilities
from .base import (
    convert_foreign_keys,
    serialize_model_instance,
    get_field_value_safely,
    is_async_context,
    get_pydantic_type,
    # Async versions
    convert_foreign_keys_async,
    serialize_model_instance_async,
    get_all_objects_async,
    get_object_or_404_async,
)

# Schema generation
from .schema import generate_schema

# Component classes
from .hooks import SyncHookExecutor, AsyncHookExecutor
from .model import SyncModelUtils, AsyncModelUtils

# Legacy functions
from .legacy import handle_response_async, execute_hook_async

__all__ = [
    # Core functions
    'convert_foreign_keys',
    'serialize_model_instance', 
    'get_field_value_safely',
    'is_async_context',
    'get_pydantic_type',
    'generate_schema',
    
    # Async versions
    'convert_foreign_keys_async',
    'serialize_model_instance_async',
    'get_all_objects_async',
    'get_object_or_404_async',
    
    # Component classes
    'SyncHookExecutor',
    'AsyncHookExecutor', 
    'SyncModelUtils',
    'AsyncModelUtils',
    
    # Legacy (deprecated)
    'handle_response_async',
    'execute_hook_async',
]



import asyncio

import pytest

from lazy_ninja.utils.hooks import (
    SyncHookExecutor,
    AsyncHookExecutor,
    execute_hook,
    execute_hook_async,
    get_hook,
)


def test_sync_hook_executor_executes_callable():
    executor = SyncHookExecutor()
    result = executor.execute(lambda value: value.upper(), "sync")
    assert result == "SYNC"


def test_async_hook_executor_executes_callable():
    executor = AsyncHookExecutor()
    result = asyncio.run(executor.execute(lambda value: value.lower(), "ASYNC"))
    assert result == "async"


def test_execute_hook_helpers(monkeypatch):
    calls = {}

    def sync_hook(value):
        calls["sync"] = value
        return value

    async_result = asyncio.run(execute_hook_async(sync_hook, "payload"))
    assert async_result == "payload"

    sync_result = execute_hook(sync_hook, "payload")
    assert sync_result == "payload"
    assert calls["sync"] == "payload"


def test_get_hook_prefers_controller_methods():
    class Controller:
        def before_create(self, value):
            return value

    controller = Controller()
    hook = get_hook(controller, "before_create")
    assert hook is not None
    assert hook("data") == "data"


def test_get_hook_falls_back_to_passed_hook():
    def fallback(value):
        return value

    class Controller:
        def before_create(self, value):
            return value

    controller = Controller()
    Controller.before_create.__is_default_hook__ = True
    hook = get_hook(controller, "before_create", passed_hook=fallback)
    assert hook is fallback



from asgiref.sync import sync_to_async, iscoroutinefunction

from django.http import HttpRequest
from django.utils.deprecation import MiddlewareMixin

class ProcessPutPatchMiddleware(MiddlewareMixin):
    """
    Middleware to handle PUT/PATCH requests as POST for form data processing
    Works for both synchronous and asynchronous request handlers
    """
    
    def _core_processing(self, request: HttpRequest) -> None:
        if request.method in ("PUT", "PATCH") and request.content_type != "application/json":
            original_method = request.method
            request.method = "POST"
            request.META["REQUEST_METHOD"] = "POST"
            
            request._load_post_and_files()
            
            request.method = original_method
            request.META["REQUEST_METHOD"] = original_method

    def __call__(self, request: HttpRequest):
        if iscoroutinefunction(self.get_response):
            async def async_handler(request):
                await sync_to_async(self._core_processing)(request)
                response = await self.get_response(request)
                return response
            return async_handler(request)
        else:
            self._core_processing(request)
            response = self.get_response(request)
            return response


import asyncio
import asyncio
import uuid
from types import SimpleNamespace

import pytest
from asgiref.sync import sync_to_async
from django.db import models
from ninja import Schema

from lazy_ninja.helpers import (
    to_kebab_case,
    parse_model_id,
    get_hook,
    execute_hook,
    handle_response,
    parse_query_param,
    apply_filters,
    apply_filters_async,
)

from tests.models import TestModel

@pytest.mark.parametrize("input_str,expected", [
    ("PostComments", "post-comments"),
    ("UserProfile", "user-profile"),
    ("APIKey", "api-key"),
    ("snake_case", "snake-case"),
    ("SCREAMING_SNAKE_CASE", "screaming-snake-case"),
    ("camelCase", "camel-case"),
    ("PascalCase", "pascal-case"),
    ("simpleword", "simpleword"),
    ("TwoWords", "two-words"),
    ("ManyManyWords", "many-many-words"),
    ("ABC", "abc"),
    ("XMLHttpRequest", "xml-http-request"),
])
def test_to_kebab_case(input_str, expected):
    """Test to_kebab_case function with various input formats"""
    result = to_kebab_case(input_str)
    assert result == expected


def test_parse_model_id_with_auto_field_returns_int():
    assert parse_model_id(TestModel, "123") == 123


def test_parse_model_id_preserves_non_numeric_strings():
    assert parse_model_id(TestModel, "abc-123") == "abc-123"


def test_parse_model_id_with_uuid_field_returns_original_value():
    class StubModel:
        class _Meta:
            pk = models.UUIDField(primary_key=True)

        _meta = _Meta()

    item_id = str(uuid.uuid4())
    assert parse_model_id(StubModel, item_id) == item_id


def test_get_hook_returns_defined_method():
    class Controller:
        def before_create(self, request, payload, schema):
            return payload

    controller = Controller()
    hook = get_hook(controller, "before_create")
    assert hook is not None
    assert hook(None, "value", None) == "value"


def test_get_hook_respects_default_hook_attribute():
    class Controller:
        def before_create(self, request, payload, schema):
            return payload

    Controller.before_create.__is_default_hook__ = True
    controller = Controller()
    hook = get_hook(controller, "before_create")
    result = execute_hook(hook, {"payload": True})
    assert result == {"payload": True}


def test_execute_hook_runs_custom_callable():
    calls = {}

    def custom_hook(value):
        calls["value"] = value
        return value.upper()

    result = execute_hook(custom_hook, "payload")
    assert result == "PAYLOAD"
    assert calls["value"] == "payload"


def test_execute_hook_returns_first_argument_for_default_hook():
    def default_hook(value):
        return "should not run"

    default_hook.__is_default_hook__ = True

    assert execute_hook(default_hook, {"key": "value"}) == {"key": "value"}


def test_handle_response_uses_custom_response():
    instance = SimpleNamespace(name="Example")

    def custom_response(request, payload):
        return {"custom": payload.name.upper()}

    result = handle_response(instance, SimpleNamespace, custom_response)
    assert result == {"custom": "EXAMPLE"}


def test_handle_response_validates_with_schema():
    class SampleSchema(Schema):
        name: str

    instance = SimpleNamespace(name="Sample")
    result = handle_response(instance, SampleSchema, None)
    assert isinstance(result, SampleSchema)
    assert result.model_dump() == {"name": "Sample"}


def test_parse_query_param_supports_multiple_operators():
    assert parse_query_param("published=true") == {"published": True}
    assert parse_query_param("count=5") == {"count": 5}
    assert parse_query_param("title=Hello") == {"title__icontains": "Hello"}
    assert parse_query_param("views>10") == {"views__gt": 10}
    assert parse_query_param("score<20") == {"score__lt": 20}
    assert parse_query_param("invalid") == {}


@pytest.mark.django_db
def test_apply_filters_filters_and_sorts_results(create_test_model, create_test_category):
    category = create_test_category(name="Filters")
    other_category = create_test_category(name="Other")
    create_test_model(title="Alpha", category=category)
    create_test_model(title="Beta", category=other_category)

    queryset = TestModel.objects.all()
    filtered = apply_filters(
        queryset,
        TestModel,
        q="title=Alpha",
        sort="title",
        order="asc",
        kwargs={"category": category, "nonexistent": "ignored"},
    )

    titles = [obj.title for obj in filtered]
    assert titles == ["Alpha"]

    sorted_qs = apply_filters(
        TestModel.objects.all(),
        TestModel,
        q=None,
        sort="title",
        order="desc",
        kwargs={},
    )
    assert [obj.title for obj in sorted_qs] == ["Beta", "Alpha"]


@pytest.mark.django_db(transaction=True)
def test_apply_filters_async_matches_sync_behavior(create_test_model, create_test_category):
    category = create_test_category(name="Async")
    create_test_model(title="Gamma", category=category)
    create_test_model(title="Delta", category=category)

    async def run():
        qs = await sync_to_async(lambda: TestModel.objects.all())()

        result = await apply_filters_async(
            qs,
            TestModel,
            q="title=Gamma",
            sort="title",
            order="asc",
            kwargs={"category": category.id},
        )
        return [item.title for item in result]

    titles = asyncio.run(run())
    assert titles == ["Gamma"]



"""
Type guards and type-safe utilities for Lazy Ninja.

This module provides type guards and helper functions to safely work with
Django's dynamic nature while keeping type checkers happy.

Type guards help avoid `type: ignore` by encapsulating complex runtime checks.
"""

import logging
from typing import Any, Dict, List, Optional, Type, TypeVar

from django.db.models import Field, Model

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Model)


# ============================================================================
# Model and Field Type Guards
# ============================================================================


def has_field(model: Type[Model], field_name: str) -> bool:
    """Check if a Django model has a specific field.
    
    Args:
        model: Django model class
        field_name: Name of the field to check
    
    Returns:
        True if field exists, False otherwise
    
    Example:
        if has_field(User, "email"):
            # Safe to access User.email
    """
    try:
        model._meta.get_field(field_name)
        return True
    except Exception:
        return False


def has_unique_field(model: Type[Model], field_name: str) -> bool:
    """Check if a Django model has a unique field.
    
    Args:
        model: Django model class
        field_name: Name of the field to check
    
    Returns:
        True if field exists and is unique, False otherwise
    
    Example:
        if has_unique_field(User, "username"):
            # Check for duplicates before creating
    """
    try:
        field = model._meta.get_field(field_name)
        return getattr(field, "unique", False)
    except Exception:
        return False


def get_model_field(model: Type[Model], field_name: str) -> Optional[Field]:
    """Safely get a field from a Django model.
    
    Args:
        model: Django model class
        field_name: Name of the field to retrieve
    
    Returns:
        Field instance if exists, None otherwise
    
    Example:
        field = get_model_field(User, "email")
        if field:
            max_length = getattr(field, "max_length", None)
    """
    try:
        return model._meta.get_field(field_name) # type: ignore
    except Exception:
        return None


def has_attribute(obj: Any, attr_name: str) -> bool:
    """Type-safe attribute check.
    
    Args:
        obj: Object to check
        attr_name: Attribute name
    
    Returns:
        True if attribute exists and is not None
    
    Example:
        if has_attribute(user, "profile"):
            user.profile.update()
    """
    return hasattr(obj, attr_name) and getattr(obj, attr_name) is not None


def get_field_choices(field: Any) -> Optional[List[Any]]:
    """Safely get choices from a Django field.
    
    Args:
        field: Django field instance
    
    Returns:
        List of choices if field has them, None otherwise
    
    Example:
        choices = get_field_choices(status_field)
        if choices:
            # Generate select options
    """
    choices = getattr(field, "choices", None)
    if choices:
        return list(choices)
    return None


def is_related_field(field: Any) -> bool:
    """Check if field is a relation (ForeignKey, ManyToMany, etc).
    
    Args:
        field: Django field instance
    
    Returns:
        True if field is a relation, False otherwise
    
    Example:
        if is_related_field(field):
            # Handle relation differently
    """
    from django.db.models.fields.related import RelatedField
    return isinstance(field, RelatedField)


def is_many_to_many(field: Any) -> bool:
    """Check if field is a ManyToMany relation.
    
    Args:
        field: Django field instance
    
    Returns:
        True if field is ManyToMany, False otherwise
    """
    from django.db.models import ManyToManyField
    return isinstance(field, ManyToManyField)


# ============================================================================
# Model Instance Type Guards
# ============================================================================


def is_model_instance(obj: Any, model: Type[T]) -> bool:
    """Check if object is an instance of a Django model.
    
    Args:
        obj: Object to check
        model: Model class to check against
    
    Returns:
        True if obj is instance of model, False otherwise
    
    Example:
        if is_model_instance(obj, User):
            # obj is a User instance
    """
    try:
        return isinstance(obj, model)
    except Exception:
        return False


def has_primary_key(obj: Any) -> bool:
    """Check if model instance has a primary key (is saved).
    
    Args:
        obj: Django model instance
    
    Returns:
        True if instance has pk (is saved), False otherwise
    
    Example:
        if has_primary_key(user):
            # User was saved to database
    """
    return hasattr(obj, "pk") and obj.pk is not None


# ============================================================================
# User Model Type Guards
# ============================================================================


def has_user_field(user_model: Type[Model], field_name: str) -> bool:
    """Check if User model has a specific field.
    
    Specialized version for User model checks.
    
    Args:
        user_model: User model class from get_user_model()
        field_name: Field name to check
    
    Returns:
        True if field exists, False otherwise
    
    Example:
        User = get_user_model()
        if has_user_field(User, "email"):
            # Can use email field
    """
    return has_field(user_model, field_name)


def has_user_manager_method(user_model: Type[Model], method_name: str) -> bool:
    """Check if User manager has a specific method.
    
    Args:
        user_model: User model class
        method_name: Manager method name to check
    
    Returns:
        True if method exists, False otherwise
    
    Example:
        if has_user_manager_method(User, "create_user"):
            User.objects.create_user(...)
    """
    return hasattr(user_model.objects, method_name)


def get_user_identifier(user: Any) -> str:
    """Get best identifier for a user (email > username > id).
    
    Args:
        user: Django user instance
    
    Returns:
        User identifier string for logging/display
    
    Example:
        logger.info("User logged in: %s", get_user_identifier(user))
    """
    email = getattr(user, "email", None)
    if email:
        return str(email)
    
    username = getattr(user, "username", None)
    if username:
        return str(username)
    
    user_id = getattr(user, "id", None) or getattr(user, "pk", None)
    if user_id:
        return f"ID:{user_id}"
    
    return "unknown"


# ============================================================================
# File Upload Type Guards
# ============================================================================


def is_uploaded_file(file: Any) -> bool:
    """Check if object is an uploaded file.
    
    Args:
        file: Object to check
    
    Returns:
        True if object has file-like attributes
    
    Example:
        if is_uploaded_file(request.FILES.get("image")):
            # Process upload
    """
    return all([
        hasattr(file, "read"),
        hasattr(file, "name"),
        has_attribute(file, "size"),
    ])


def get_file_content_type(file: Any) -> Optional[str]:
    """Safely get content type from uploaded file.
    
    Args:
        file: Uploaded file object
    
    Returns:
        Content type string or None
    
    Example:
        content_type = get_file_content_type(file)
        if content_type and content_type.startswith("image/"):
            # Process image
    """
    return getattr(file, "content_type", None)


def get_file_size(file: Any) -> int:
    """Safely get size from uploaded file.
    
    Args:
        file: Uploaded file object
    
    Returns:
        File size in bytes, 0 if not available
    
    Example:
        size = get_file_size(file)
        if size > MAX_SIZE:
            raise ValidationError("File too large")
    """
    return getattr(file, "size", 0)


def is_image_file(file: Any) -> bool:
    """Check if uploaded file is an image.
    
    Args:
        file: Uploaded file object
    
    Returns:
        True if file is an image based on content type
    
    Example:
        if is_image_file(file):
            # Process as image
    """
    content_type = get_file_content_type(file)
    return bool(content_type and content_type.startswith("image/"))


def is_video_file(file: Any) -> bool:
    """Check if uploaded file is a video.
    
    Args:
        file: Uploaded file object
    
    Returns:
        True if file is a video based on content type
    
    Example:
        if is_video_file(file):
            # Process as video
    """
    content_type = get_file_content_type(file)
    return bool(content_type and content_type.startswith("video/"))


# ============================================================================
# Query and Filter Type Guards
# ============================================================================


def is_valid_filter_field(model: Type[Model], field_name: str) -> bool:
    """Check if field name is valid for filtering.
    
    Handles Django lookups like "field__icontains".
    
    Args:
        model: Django model class
        field_name: Field name with optional lookup
    
    Returns:
        True if field exists and is filterable
    
    Example:
        if is_valid_filter_field(User, "email__iexact"):
            # Safe to use in filter
    """
    # Handle Django lookups (field__lookup)
    base_field = field_name.split("__")[0]
    return has_field(model, base_field)


def sanitize_filters(
    model: Type[Model], 
    filters: Dict[str, Any]
) -> Dict[str, Any]:
    """Sanitize filter dict to only include valid model fields.
    
    Args:
        model: Django model class
        filters: Raw filter dictionary
    
    Returns:
        Sanitized filter dict with only valid fields
    
    Example:
        safe_filters = sanitize_filters(User, request.GET.dict())
        queryset = User.objects.filter(**safe_filters)
    """
    sanitized = {}
    for field_name, value in filters.items():
        if is_valid_filter_field(model, field_name):
            sanitized[field_name] = value
        else:
            logger.warning(
                "Ignoring invalid filter field: %s for model %s",
                field_name,
                model.__name__
            )
    return sanitized


def get_model_field_names(
    model: Type[Model],
    exclude_relations: bool = False,
) -> List[str]:
    """Get list of all field names for a model.
    
    Args:
        model: Django model class
        exclude_relations: If True, exclude ForeignKey/ManyToMany fields
    
    Returns:
        List of field names
    
    Example:
        fields = get_model_field_names(User)
        # ["id", "username", "email", ...]
    """
    try:
        fields = model._meta.get_fields()
        field_names = []
        
        for field in fields:
            if exclude_relations and is_related_field(field):
                continue
            field_name = getattr(field, "name", None)
            if field_name:
                field_names.append(field_name)
        
        return field_names
    except Exception as e:
        logger.error("Error getting field names for %s: %s", model.__name__, e)
        return []


# ============================================================================
# Validation Type Guards
# ============================================================================


def is_valid_pagination_params(page: Any, page_size: Any) -> bool:
    """Check if pagination parameters are valid.
    
    Args:
        page: Page number
        page_size: Items per page
    
    Returns:
        True if both are valid integers
    
    Example:
        if is_valid_pagination_params(request.GET.get("page"), ...):
            # Safe to paginate
    """
    try:
        p = int(page)
        ps = int(page_size)
        return p >= 1 and 1 <= ps <= 100
    except (ValueError, TypeError):
        return False


def is_uuid(value: Any) -> bool:
    """Check if value is a valid UUID.
    
    Args:
        value: Value to check
    
    Returns:
        True if value is a valid UUID
    
    Example:
        if is_uuid(user_id):
            # Safe to use as UUID
    """
    import uuid
    try:
        uuid.UUID(str(value))
        return True
    except (ValueError, TypeError, AttributeError):
        return False



"""Route registration facade delegating to sync/async model routers."""
from typing import Type, Optional, Callable, Any

from django.db.models import Model

from ninja import NinjaAPI, Schema
from pydantic import BaseModel

from .router.base import BaseModelRouter
from .router.async_router import AsyncModelRouter
from .router.sync_router import SyncModelRouter
from .pagination import BasePagination
from .file_upload import FileUploadConfig


def register_model_routes_internal(
    api: NinjaAPI,
    model: Type[Model],
    base_url: str,
    list_schema: Type[BaseModel],
    detail_schema: Type[BaseModel],
    create_schema: Optional[Type[BaseModel]] = None,
    update_schema: Optional[Type[BaseModel]] = None,
    pre_list: Optional[Callable[[Any, Any], Any]] = None,
    before_create: Optional[Callable[[Any, Any, Type[Schema]], Any]] = None,
    after_create: Optional[Callable[[Any, Any], Any]] = None,
    before_update: Optional[Callable[[Any, Any, Type[Schema]], Any]] = None,
    after_update: Optional[Callable[[Any, Any], Any]] = None,
    before_delete: Optional[Callable[[Any, Any], None]] = None,
    after_delete: Optional[Callable[[Any], None]] = None,
    custom_response: Optional[Callable[[Any, Any], Any]] = None,
    pagination_strategy: Optional[BasePagination] = None,
    file_upload_config: Optional[FileUploadConfig] = None,
    use_multipart_create: bool = False,
    use_multipart_update: bool = False,
    is_async: bool = True,
) -> None:
    """Register CRUD routes for a Django model using the appropriate router implementation."""

    router_cls: Type[BaseModelRouter]
    router_cls = AsyncModelRouter if is_async else SyncModelRouter

    router = router_cls(
        api=api,
        model=model,
        base_url=base_url,
        list_schema=list_schema,
        detail_schema=detail_schema,
        create_schema=create_schema,
        update_schema=update_schema,
        pagination_strategy=pagination_strategy,
        file_upload_config=file_upload_config,
        use_multipart_create=use_multipart_create,
        use_multipart_update=use_multipart_update,
        pre_list=pre_list,
        before_create=before_create,
        after_create=after_create,
        before_update=before_update,
        after_update=after_update,
        before_delete=before_delete,
        after_delete=after_delete,
        custom_response=custom_response,
    )

    router.finalize()



"""Auth hook helpers."""

from typing import Any, Callable, Optional

from .config import get_auth_hook


def on_login_hook() -> Optional[Callable[..., Any]]:
    return get_auth_hook(["ON_LOGIN", "LOGIN_HOOK"])


def on_register_hook() -> Optional[Callable[..., Any]]:
    return get_auth_hook(["ON_REGISTER", "REGISTER_HOOK"])


def on_refresh_hook() -> Optional[Callable[..., Any]]:
    return get_auth_hook(["ON_REFRESH", "REFRESH_HOOK"])


def on_logout_hook() -> Optional[Callable[..., Any]]:
    return get_auth_hook(["ON_LOGOUT", "LOGOUT_HOOK"])



import pytest
from django.test import Client

from lazy_ninja.builder import ExclusionConfig

@pytest.mark.django_db
def test_dynamic_api_registration(client):
    """Tests if all routes are registered correctly"""
    client = Client()
    url = "/api/test-models/"
    response = client.get(url)
    assert response.status_code == 200
    assert response.json() ==  {
        'count': 0,
        'items': []
    }


def make_model(app_label, name):
    return type(name, (), {"_meta": type("_meta", (), {"app_label": app_label})()})


def test_exclusion_config_default_system_apps_excluded():
    config = ExclusionConfig()
    auth_model = make_model("auth", "User")
    assert config.should_exclude_model(auth_model) is True


def test_exclusion_config_custom_rules():
    model_a = make_model("blog", "Post")
    model_b = make_model("blog", "Comment")
    config = ExclusionConfig(exclude={"blog": {"Comment"}})

    assert config.should_exclude_model(model_a) is False
    assert config.should_exclude_model(model_b) is True

    config = ExclusionConfig(exclude={"shop": True})
    shop_model = make_model("shop", "Order")
    assert config.should_exclude_model(shop_model) is True



from typing import List, Any, Dict, Optional, Union

from django.shortcuts import get_object_or_404
from django.db.models import QuerySet
from ninja import Form
from ninja.pagination import paginate

from .base import BaseModelRouter
from ..handlers.response import SyncResponseHandler
from ..handlers.file_handler import SyncFileHandler
from ..utils.hooks import SyncHookExecutor
from ..utils.model import SyncModelUtils
from ..helpers import QuerysetFilter, parse_model_id
from ..errors import handle_exception


class SyncModelRouter(BaseModelRouter):
    """
    Sync implementation of model router.
    
    Handles all sync route registration and request processing.
    """
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.response_handler = SyncResponseHandler()
        self.file_handler = SyncFileHandler(self.file_upload_config)
        self.hook_executor = SyncHookExecutor()
        self.model_utils = SyncModelUtils()
        self.queryset_filter = QuerysetFilter(self.model)
    
    def register_list_route(self) -> None:
        """Register sync list route with pagination and filtering."""
        
        @self.router.get(
            "/", 
            response=List[self.list_schema], 
            tags=self.get_tags(), 
            operation_id=self.get_operation_id("list")
        )
        @paginate(self.paginator_class)
        def list_items(
            request, 
            q: Optional[str] = None, 
            sort: Optional[str] = None,
            order: Optional[str] = "asc", 
            **kwargs: Any
        ) -> Union[QuerySet, Any]:
            """List objects with optional filtering and sorting."""
            try:
                queryset = self.model.objects.all()
                
                if self.pre_list:
                    queryset = self.hook_executor.execute(self.pre_list, request, queryset) or queryset
                
                queryset = self.queryset_filter.apply_filters(queryset, q, sort, order, **kwargs)
                
                return queryset if not self.custom_response else self.custom_response(request, queryset)
            except Exception as e:
                return handle_exception(e)
    
    def register_detail_route(self) -> None:
        """Register sync detail route."""
        
        @self.router.get(
            "/{item_id}", 
            response=self.detail_schema, 
            tags=self.get_tags(), 
            operation_id=self.get_operation_id("get")
        )
        def get_item(request, item_id: str) -> Any:
            """Retrieve a single object by ID."""
            try:
                item_id_value = parse_model_id(self.model, item_id)
                instance = get_object_or_404(self.model, id=item_id_value)
                return self.response_handler.handle_response(
                    instance, self.detail_schema, self.custom_response, request
                )
            except Exception as e:
                return handle_exception(e)
    
    def register_create_route(self) -> None:
        """Register sync create route."""
        if self.use_multipart_create:
            self._register_multipart_create_route()
        else:
            self._register_json_create_route()
    
    def _register_json_create_route(self) -> None:
        """Register JSON-based create route."""
        
        @self.router.post(
            "/", 
            response=self.detail_schema, 
            tags=self.get_tags(), 
            operation_id=self.get_operation_id("create")
        )
        def create_item(request, payload: self.create_schema) -> Any: # type: ignore
            """Create a new object."""
            try:
                if self.before_create:
                    payload = self.hook_executor.execute(
                        self.before_create, request, payload, self.create_schema
                    ) or payload
                
                data = self.model_utils.convert_foreign_keys(self.model, payload.model_dump())
                
                instance = self.model.objects.create(**data)
                
                if self.after_create:
                    instance = self.hook_executor.execute(
                        self.after_create, request, instance
                    ) or instance
                
                return self.response_handler.handle_response(
                    instance, self.detail_schema, self.custom_response, request
                )
            except Exception as e:
                return handle_exception(e)
    
    def _register_multipart_create_route(self) -> None:
        """Register multipart/form-data create route."""
        
        @self.router.post(
            "/", 
            response=self.detail_schema, 
            tags=self.get_tags(), 
            operation_id=self.get_operation_id("create")
        )
        def create_item(request, payload: self.create_schema = Form(...)) -> Any: # type: ignore
            """Create a new object with file upload support."""
            try:
                if self.before_create:
                    payload = self.hook_executor.execute(
                        self.before_create, request, payload, self.create_schema
                    ) or payload
                
                data, file_fields_map = self.file_handler.process_create_files(
                    request, payload, self.model
                )
                
                data = self.model_utils.convert_foreign_keys(self.model, data)
                
                instance = self.model.objects.create(**data)
                
                if file_fields_map:
                    self.file_handler.handle_file_relations(
                        instance, file_fields_map, self.model
                    )
                
                if self.after_create:
                    instance = self.hook_executor.execute(
                        self.after_create, request, instance
                    ) or instance
                
                return self.response_handler.handle_response(
                    instance, self.detail_schema, self.custom_response, request
                )
            except Exception as e:
                return handle_exception(e)
    
    def register_update_route(self) -> None:
        """Register sync update route."""
        if self.use_multipart_update:
            self._register_multipart_update_route()
        else:
            self._register_json_update_route()
    
    def _register_json_update_route(self) -> None:
        """Register JSON-based update route."""
        
        @self.router.patch(
            "/{item_id}", 
            response=self.detail_schema, 
            tags=self.get_tags(), 
            operation_id=self.get_operation_id("update")
        )
        def update_item(request, item_id: str, payload: self.update_schema) -> Any: # type: ignore
            """Update an existing object."""
            try:
                item_id_value = parse_model_id(self.model, item_id)
                instance = get_object_or_404(self.model, id=item_id_value)

                if self.before_update:
                    payload = self.hook_executor.execute(
                        self.before_update, request, instance, payload, self.update_schema
                    ) or payload
                
                data = self.model_utils.convert_foreign_keys(
                    self.model, payload.model_dump(exclude_unset=True)
                )
                
                for key, value in data.items():
                    setattr(instance, key, value)
                instance.save()
                
                if self.after_update:
                    instance = self.hook_executor.execute(
                        self.after_update, request, instance
                    ) or instance
                
                return self.response_handler.handle_response(
                    instance, self.detail_schema, self.custom_response, request
                )
            except Exception as e:
                return handle_exception(e)
    
    def _register_multipart_update_route(self) -> None:
        """Register multipart/form-data update route."""
        
        @self.router.patch(
            "/{item_id}", 
            response=self.detail_schema, 
            tags=self.get_tags(), 
            operation_id=self.get_operation_id("update")
        )
        def update_item(request, item_id: str, payload: self.update_schema = Form(...)) -> Any: # type: ignore
            """Update an existing object with file upload support."""
            try:
                item_id_value = parse_model_id(self.model, item_id)
                instance = get_object_or_404(self.model, id=item_id_value)
                
                if self.before_update:
                    payload = self.hook_executor.execute(
                        self.before_update, request, instance, payload, self.update_schema
                    ) or payload
                
                data, file_fields_map = self.file_handler.process_update_files(
                    request, payload, self.model
                )
                
                data = self.model_utils.convert_foreign_keys(self.model, data)
                
                for key, value in data.items():
                    setattr(instance, key, value)
                instance.save()
                
                if file_fields_map:
                    self.file_handler.handle_file_relations(
                        instance, file_fields_map, self.model
                    )
                
                if self.after_update:
                    instance = self.hook_executor.execute(
                        self.after_update, request, instance
                    ) or instance
                
                return self.response_handler.handle_response(
                    instance, self.detail_schema, self.custom_response, request
                )
            except Exception as e:
                return handle_exception(e)
    
    def register_delete_route(self) -> None:
        """Register sync delete route."""
        
        @self.router.delete(
            "/{item_id}", 
            response={200: Dict[str, str]}, 
            tags=self.get_tags(), 
            operation_id=self.get_operation_id("delete")
        )
        def delete_item(request, item_id: str) -> Dict[str, str]:
            """Delete an object."""
            try:
                item_id_value = parse_model_id(self.model, item_id)
                instance = get_object_or_404(self.model, id=item_id_value)

                if self.before_delete:
                    self.hook_executor.execute(self.before_delete, request, instance)
                
                instance.delete()
                
                if self.after_delete:
                    self.hook_executor.execute(self.after_delete, instance)
                
                return {"message": f"{self.model.__name__} with ID {item_id} has been deleted."}
            except Exception as e:
                return handle_exception(e)



import asyncio
import json

import pytest
from django.http import HttpResponse
from django.test import RequestFactory
from django.test.client import BOUNDARY, MULTIPART_CONTENT, encode_multipart

from lazy_ninja.errors import LazyNinjaError
from lazy_ninja.middleware.error_handling import ErrorHandlingMiddleware
from lazy_ninja.middleware.process_put_patch import ProcessPutPatchMiddleware


def test_error_handling_middleware_passes_through_response():
    factory = RequestFactory()
    request = factory.get("/")

    middleware = ErrorHandlingMiddleware(lambda req: HttpResponse("ok"))
    response = middleware(request)
    assert response.status_code == 200
    assert response.content == b"ok"


def test_error_handling_middleware_handles_lazy_error():
    factory = RequestFactory()
    request = factory.get("/")

    def view(_):
        raise LazyNinjaError("boom", status_code=418)

    middleware = ErrorHandlingMiddleware(view)
    response = middleware(request)
    assert response.status_code == 418
    payload = json.loads(response.content.decode())
    assert payload["error"]["message"] == "boom"


def test_error_handling_middleware_wraps_generic_exception():
    factory = RequestFactory()
    request = factory.get("/")

    def view(_):
        raise RuntimeError("unexpected")

    middleware = ErrorHandlingMiddleware(view)
    response = middleware(request)
    assert response.status_code == 500
    payload = json.loads(response.content.decode())
    assert payload["error"]["type"] == "LazyNinjaError"


def test_error_handling_middleware_async_call():
    factory = RequestFactory()
    request = factory.get("/")

    async def view(_):
        raise LazyNinjaError("async", status_code=409)

    middleware = ErrorHandlingMiddleware(view)
    response = asyncio.run(middleware.__acall__(request))
    assert response.status_code == 409


def test_process_put_patch_middleware_sync_put_request():
    factory = RequestFactory()
    body = encode_multipart(BOUNDARY, {"name": "sync"})
    request = factory.put("/", data=body, content_type=MULTIPART_CONTENT)

    middleware = ProcessPutPatchMiddleware(lambda req: HttpResponse("ok"))
    response = middleware(request)

    assert response.status_code == 200
    assert request.method == "PUT"
    assert request.POST["name"] == "sync"


def test_process_put_patch_middleware_async_patch_request():
    factory = RequestFactory()
    body = encode_multipart(BOUNDARY, {"name": "async"})
    request = factory.patch("/", data=body, content_type=MULTIPART_CONTENT)

    async def view(req):
        assert req.method == "PATCH"
        return HttpResponse("async")

    middleware = ProcessPutPatchMiddleware(view)
    response = asyncio.run(middleware(request))

    assert response.status_code == 200
    assert request.POST["name"] == "async"



"""Auth configuration helpers."""

from typing import Any, Dict, Optional, Callable

from django.conf import settings


def _auth_cfg() -> Dict[str, Any]:
    return getattr(settings, "LAZY_NINJA_AUTH", {}) or {}


def get_setting(keys: list[str], default: Any = None) -> Any:
    """Resolve setting from LAZY_NINJA_AUTH or Django settings."""
    cfg = _auth_cfg()
    for key in keys:
        if key in cfg:
            return cfg[key]
        val = getattr(settings, key, None)
        if val is not None:
            return val
    return default


def get_jwt_secret() -> str:
    """Return the secret key used to sign JWTs."""
    secret = get_setting(["JWT_SECRET", "SECRET_KEY"])
    if not secret:
        raise RuntimeError("JWT secret is not configured.")
    return str(secret)


def get_jwt_algorithm() -> str:
    return get_setting(["JWT_ALGORITHM"], "HS256")


def get_jwt_issuer() -> str:
    return str(get_setting(["JWT_ISS", "JWT_ISSUER"], "lazy-ninja"))


def get_jwt_audience() -> str:
    return str(get_setting(["JWT_AUD", "JWT_AUDIENCE"], "lazy-ninja-api"))


def get_token_lifetimes() -> Dict[str, int]:
    """Return access and refresh token lifetimes in seconds."""
    return {
        "access": int(get_setting(["JWT_ACCESS_EXP"], 60 * 60 * 24)),
        "refresh": int(get_setting(["JWT_REFRESH_EXP"], 60 * 60 * 24 * 30)),
    }


def cookie_secure_flag() -> bool:
    return bool(get_setting(["COOKIE_SECURE"], not getattr(settings, "DEBUG", False)))


def should_validate_password() -> bool:
    """Check if password validation is enabled (default: True)."""
    return bool(get_setting(["VALIDATE_PASSWORD"], True))


def should_log_auth_events() -> bool:
    """Check if authentication events should be logged (default: True)."""
    return bool(get_setting(["LOG_AUTH_EVENTS"], True))


def is_stateful() -> bool:
    """Check if stateful token mode is enabled (default: False)."""
    return bool(get_setting(["STATEFUL"], False))


def get_blacklist_prefix() -> str:
    return str(get_setting(["BLACKLIST_PREFIX"], "lazy_ninja:bl"))


def should_set_cookies() -> bool:
    """Check if auth endpoints should set cookies (default: True)."""
    return bool(get_setting(["SET_COOKIES"], True))


def should_rotate_refresh() -> bool:
    """Check if refresh tokens should be rotated (default: True)."""
    return bool(get_setting(["ROTATE_REFRESH"], True))


def get_auth_hook(keys: list[str]) -> Optional[Callable[..., Any]]:
    hook = get_setting(keys)
    return hook if callable(hook) else None


def get_login_fields() -> list[str]:
    login_fields_setting = get_setting(["LOGIN_FIELDS"], ["username"])
    if isinstance(login_fields_setting, (list, tuple, set)):
        return [str(field) for field in login_fields_setting]
    return [str(login_fields_setting)]



import asyncio
import json

import pytest
from django.core import exceptions
from django.db import DatabaseError
from django.http import JsonResponse

from ninja.errors import HttpError

from lazy_ninja.errors import (
    handle_exception,
    handle_exception_async,
    LazyNinjaError,
    ValidationError,
    NotFoundError,
    PermissionDeniedError,
    DatabaseOperationError,
    SynchronousOperationError,
)

from tests.models import TestModel


def _extract_payload(response: JsonResponse) -> dict:
    assert response.status_code
    data = json.loads(response.content.decode())
    assert "error" in data
    return data["error"]


@pytest.mark.parametrize(
    "exception,expected_type,expected_status",
    [
        (TestModel.DoesNotExist("missing"), NotFoundError, 404),
        (exceptions.ObjectDoesNotExist("missing"), NotFoundError, 404),
        (PermissionError("denied"), PermissionDeniedError, 403),
        (exceptions.ValidationError("invalid"), ValidationError, 400),
        (ValueError("bad"), ValidationError, 400),
        (DatabaseError("db error"), DatabaseOperationError, 500),
        (exceptions.SynchronousOnlyOperation("sync"), SynchronousOperationError, 500),
        (HttpError(401, "unauthorized"), LazyNinjaError, 401),
    ],
)
def test_handle_exception_maps_common_errors(exception, expected_type, expected_status, settings):
    settings.DEBUG = False
    response = handle_exception(exception)
    assert response.status_code == expected_status
    error = _extract_payload(response)
    assert error["type"] == expected_type.__name__


def test_handle_exception_preserves_lazy_ninja_error():
    error = LazyNinjaError("boom", status_code=499)
    response = handle_exception(error)
    assert response.status_code == 499
    payload = _extract_payload(response)
    assert payload["message"] == "boom"
    assert payload["type"] == "LazyNinjaError"


def test_handle_exception_defaults_to_generic_error(settings, capsys):
    settings.DEBUG = True
    response = handle_exception(RuntimeError("crash"))
    assert response.status_code == 500
    payload = _extract_payload(response)
    assert payload["type"] == "LazyNinjaError"
    captured = capsys.readouterr().out
    assert "LazyNinja" in captured


def test_handle_exception_async_delegates_to_sync():
    async def run():
        response = await handle_exception_async(TestModel.DoesNotExist("async"))
        return response.status_code

    status = asyncio.run(run())
    assert status == 404




from __future__ import annotations

import subprocess
import sys

import pytest

from lazy_ninja.cli import startproject as sp


def test_find_project_package_nested_layout(tmp_path):
    """
    layout:
    tmp/project_name/project_name/settings.py  -> should return tmp/project_name/project_name
    """
    root = tmp_path
    project_name = "myproj"
    p = root / project_name / project_name
    p.mkdir(parents=True)
    settings = p / "settings.py"
    settings.write_text("# settings")
    found = sp._find_project_package(root, project_name)
    assert found == p


def test_find_project_package_flat_layout(tmp_path):
    """
    layout:
    tmp/project_name/settings.py  -> should return tmp/project_name
    """
    root = tmp_path
    project_name = "myproj"
    p = root / project_name
    p.mkdir(parents=True)
    settings = p / "settings.py"
    settings.write_text("# settings")
    found = sp._find_project_package(root, project_name)
    assert found == p


def test_find_project_package_root_settings(tmp_path):
    """
    layout:
    tmp/settings.py -> should return tmp
    """
    root = tmp_path
    settings = root / "settings.py"
    settings.write_text("# settings")
    found = sp._find_project_package(root, "anyname")
    assert found == root


def test_find_project_package_multiple_matches_prefers_project_name(tmp_path):
    """
    Create two settings.py files; one whose parent matches project_name should be chosen.
    """
    root = tmp_path
    (root / "a").mkdir()
    (root / "a" / "settings.py").write_text("# a settings")
    (root / "myproj").mkdir()
    (root / "myproj" / "settings.py").write_text("# myproj settings")

    found = sp._find_project_package(root, "myproj")
    assert found.name == "myproj"


def test_find_project_package_no_matches_raises(tmp_path):
    with pytest.raises(RuntimeError):
        sp._find_project_package(tmp_path, "nonexistent")


def test_run_django_startproject_uses_django_admin_when_available(monkeypatch):
    calls = {}

    monkeypatch.setattr(sp.shutil, "which", lambda name: "/usr/bin/django-admin")

    def fake_run(cmd, check=False):
        calls["cmd"] = cmd
        calls["check"] = check
        class R: pass
        return R()

    monkeypatch.setattr(sp.subprocess, "run", fake_run)

    sp._run_django_startproject("myproj", "targetdir")
    assert calls["cmd"][0] == "django-admin"
    assert calls["cmd"][1] == "startproject"
    assert calls["cmd"][2] == "myproj"
    assert "targetdir" in calls["cmd"]


def test_run_django_startproject_falls_back_to_python_module(monkeypatch):
    calls = {}
    monkeypatch.setattr(sp.shutil, "which", lambda name: None)

    def fake_run(cmd, check=False):
        calls["cmd"] = cmd
        calls["check"] = check
        class R: pass
        return R()

    monkeypatch.setattr(sp.subprocess, "run", fake_run)

    sp._run_django_startproject("myproj", "targetdir")
    assert calls["cmd"][0] == sys.executable
    assert calls["cmd"][1] == "-m"
    assert calls["cmd"][2] == "django"
    assert "startproject" in calls["cmd"]


def test_scaffold_creates_api_and_appends_settings(tmp_path, capsys):
    root = tmp_path
    project_name = "myproj"
    pkg = root / project_name / project_name
    pkg.mkdir(parents=True)
    settings_file = pkg / "settings.py"
    settings_file.write_text("# base settings\n")

    sp._scaffold_lazy_ninja(root, project_name, title="My API Title")

    api = pkg / "api.py"
    assert api.exists()
    content = api.read_text()
    assert "NinjaAPI" in content
    assert "My API Title" in content

    settings_text = settings_file.read_text()
    assert "Lazy Ninja scaffold" in settings_text


def test_scaffold_skips_existing_api(tmp_path, capsys):
    root = tmp_path
    project_name = "myproj"
    pkg = root / project_name / project_name
    pkg.mkdir(parents=True)
    settings_file = pkg / "settings.py"
    settings_file.write_text("# settings\n")

    api_file = pkg / "api.py"
    api_file.write_text("# existing api\n")

    sp._scaffold_lazy_ninja(root, project_name, title=None)
    assert api_file.read_text() == "# existing api\n"


def test_startproject_command_handles_django_start_failure(monkeypatch):
    def raise_called(name, target=None):
        raise subprocess.CalledProcessError(returncode=3, cmd="failcmd")

    monkeypatch.setattr(sp, "_run_django_startproject", raise_called)

    with pytest.raises(SystemExit) as exc:
        sp.startproject_command("myproj", directory="targetdir", title=None)
    assert exc.value.code == 3


def test_startproject_command_scaffold_failure_prints_warning(monkeypatch, tmp_path, capsys):
    monkeypatch.setattr(sp, "_run_django_startproject", lambda name, target=None: None)
    def raise_exc(root, name, title):
        raise RuntimeError("boom")
    monkeypatch.setattr(sp, "_scaffold_lazy_ninja", raise_exc)

    sp.startproject_command("myproj", directory=str(tmp_path), title="T")

    out = capsys.readouterr().out
    assert "Project created but failed to add Lazy Ninja scaffold" in out or "failed to add Lazy Ninja scaffold" in out



from typing import Type, Any, Dict
from asgiref.sync import sync_to_async

from django.db import models
from django.shortcuts import get_object_or_404

from .base import serialize_model_instance, serialize_model_instance_async

class BaseModelUtils:
    """Base class for model utilities."""

    def convert_foreign_keys(self, model: Type[models.Model], data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Converts integer values for ForeignKey fields in `data` to the corresponding model instances.
        
        Args:
            model: The Django model class
            data: Dictionary containing field data
            
        Returns:
            Dictionary with converted foreign key values
        """
        for field in model._meta.fields:
            if isinstance(field, models.ForeignKey) and field.name in data:
                fk_value = data[field.name]
                if isinstance(fk_value, (int, str)):
                    data[field.name] = field.related_model.objects.get(pk=fk_value)
        return data


class SyncModelUtils(BaseModelUtils):
    """Handles model operations for sync routes."""
    
    def get_object_or_404(self, model: Type[models.Model], **kwargs) -> Any:
        """Get object or raise 404."""
        return get_object_or_404(model, **kwargs)
    
    def create_instance(self, model: Type[models.Model], **data) -> Any:
        """Create a new model instance."""
        return model.objects.create(**data)
    
    def update_instance(self, instance: Any, data: Dict[str, Any]) -> None:
        """Update an existing model instance."""
        for key, value in data.items():
            setattr(instance, key, value)
        instance.save()
    
    def delete_instance(self, instance: Any) -> None:
        """Delete a model instance."""
        instance.delete()
    
    def serialize_model_instance(self, instance: Any) -> Dict[str, Any]:
        """Serialize a model instance."""
        return serialize_model_instance(instance)
    

class AsyncModelUtils(BaseModelUtils):
    """Handles model operations for async routes."""

    async def get_all_objects(self, model: Type[models.Model]):
        """Get all objects for a model asynchronously."""
        return await sync_to_async(lambda m: m.objects.all())(model)
    
    async def get_object_or_404(self, model: Type[models.Model], **kwargs) -> Any:
        """Get object or raise 404 asynchronously."""
        return await sync_to_async(get_object_or_404)(model, **kwargs)
    
    async def create_instance(self, model: Type[models.Model], **data) -> Any:
        """Create a new model instance asynchronously."""
        create_func = sync_to_async(lambda m, **kwargs: m.objects.create(**kwargs))
        return await create_func(model, **data)
    
    async def update_instance(self, instance: Any, data: Dict[str, Any]) -> None:
        """Update an existing model instance asynchronously."""
        for key, value in data.items():
            setattr(instance, key, value)

        save_instance = sync_to_async(lambda obj: obj.save())
        await save_instance(instance)

    async def delete_instance(self, instance: Any) -> None:
        """Delete a model instance asynchronously."""
        delete_func = sync_to_async(lambda obj: obj.delete())
        await delete_func(instance)

    async def convert_foreign_keys(self, model: Type[models.Model], data: Dict[str, Any]) -> Dict[str, Any]:
        """Convert foreign keys asynchronously."""
        return await sync_to_async(super().convert_foreign_keys)(model, data)

    async def serialize_model_instance(self, instance: Any) -> Dict[str, Any]:
        """Serialize a model instance asynchronously."""
        return await serialize_model_instance_async(instance)
    
# Legacy function wrappers for backward compatibility
def convert_foreign_keys(model: Type[models.Model], data: Dict[str, Any]) -> Dict[str, Any]:
    """Legacy wrapper for foreign key conversion."""
    utils = SyncModelUtils()
    return utils.convert_foreign_keys(model, data)


async def convert_foreign_keys_async(model: Type[models.Model], data: Dict[str, Any]) -> Dict[str, Any]:
    """Legacy wrapper for async foreign key conversion."""
    utils = AsyncModelUtils()
    return await utils.convert_foreign_keys(model, data)



from abc import ABC, abstractmethod
from typing import Type, Optional, List, Any

from django.db.models import Model
from ninja import Router, NinjaAPI
from pydantic import BaseModel

from ..pagination import BasePagination
from ..file_upload import FileUploadConfig


class BaseModelRouter(ABC):
    """
    Base class for model routers that handles shared configuration and wiring.

    Subclasses implement the actual route registration logic for sync/async patterns.
    """

    def __init__(
        self,
        api: NinjaAPI,
        model: Type[Model],
        base_url: str,
        list_schema: Type[BaseModel],
        detail_schema: Type[BaseModel],
        create_schema: Optional[Type[BaseModel]] = None,
        update_schema: Optional[Type[BaseModel]] = None,
        pagination_strategy: Optional[BasePagination] = None,
        file_upload_config: Optional[FileUploadConfig] = None,
        use_multipart_create: bool = False,
        use_multipart_update: bool = False,
        controller: Optional[Any] = None,
        **hooks
    ):
        """
        Initialize the base router with shared configuration.

        Args:
            api: NinjaAPI instance
            model: Django model class
            base_url: Base URL for the routes
            list_schema: Schema for list responses
            detail_schema: Schema for detail responses
            create_schema: Schema for create requests
            update_schema: Schema for update requests
            pagination_strategy: Strategy for pagination
            file_upload_config: Configuration for file uploads
            use_multipart_create: Whether to use multipart for create
            use_multipart_update: Whether to use multipart for update
            controller:
            **hooks: Hook functions (before_create, pre_list, etc.)
        """
        self.api = api
        self.model = model
        self.base_url = base_url
        self.list_schema = list_schema
        self.detail_schema = detail_schema
        self.create_schema = create_schema
        self.update_schema = update_schema
        self.pagination_strategy = pagination_strategy
        self.file_upload_config = file_upload_config
        self.use_multipart_create = use_multipart_create
        self.use_multipart_update = use_multipart_update
        self.controller = controller

        self.pre_list = hooks.get('pre_list')
        self.before_create = hooks.get('before_create')
        self.after_create = hooks.get('after_create')
        self.before_update = hooks.get('before_update')
        self.after_update = hooks.get('after_update')
        self.before_delete = hooks.get('before_delete')
        self.after_delete = hooks.get('after_delete')
        self.custom_response = hooks.get('custom_response')

        self.model_name = model.__name__.lower()
        self.paginator_class = pagination_strategy.get_paginator() if pagination_strategy else None

        self.router = Router()

    @abstractmethod
    def register_list_route(self) -> None:
        """Register the list route."""
        pass

    @abstractmethod
    def register_detail_route(self) -> None:
        """Register the detail route."""
        pass

    @abstractmethod
    def register_create_route(self) -> None:
        """Register the create route."""
        pass

    @abstractmethod
    def register_update_route(self) -> None:
        """Register the update route."""
        pass

    @abstractmethod
    def register_delete_route(self) -> None:
        """Register the delete route."""
        pass

    
    def finalize(self) -> None:
        """
        Register all routes and add the router to the API.

        This method orchestrates the route registration process.
        """
        self.register_list_route()
        self.register_detail_route()

        if self.create_schema:
            self.register_create_route()

        if self.update_schema:
            self.register_update_route()
        
        self.register_delete_route()

        self.api.add_router(self.base_url, self.router)

    def get_operation_id(self, operation: str) -> str:
        """Generate operation ID for OpenAPI"""
        return f'{operation}_{self.model_name}'
    
    def get_tags(self) -> List[str]:
        """Get tags for route grouping."""
        return [self.model.__name__]
    



from typing import Dict, List, Tuple

from django.db import models


class FileUploadConfig:
    """
    Configuration for file upload fields in a model.
    
    This class helps configure which fields are file fields and
    whether to use multipart/form-data for specific models.
    """
    def __init__(
        self,
        file_fields: Dict[str, List[str]] = None,
        multiple_file_fields: Dict[str, List[str]] = None,
    ):
        """
        Initialize file upload configuration.
        
        Args:
            file_fields: Dictionary mapping model names to lists of file field names
                         e.g. {"MyModel": ["image", "attachment"]}
            multiple_file_fields: Dictionary mapping model names to lists of field names 
                                  that can accept multiple files
                                  e.g. {"MyModel": ["gallery_images"]}
        """
        self.file_fields = file_fields or {}
        self.multiple_file_fields = multiple_file_fields or {}
        
    def get_model_file_fields(self, model_name: str) -> List[str]:
        """Get list of file fields for a model."""
        return self.file_fields.get(model_name, [])
    
    def get_model_multiple_file_fields(self, model_name: str) -> List[str]:
        """Get list of multiple file fields for a model."""
        return self.multiple_file_fields.get(model_name, [])
    
    def is_multiple_file_field(self, model_name: str, field_name: str) -> bool:
        """Check if a field is configured for multiple file uploads."""
        return field_name in self.get_model_multiple_file_fields(model_name)


class FileFieldDetector:
    """
    Utility class for detecting file fields in Django models.
    
    Separated from the main detect_file_fields function for better organization.
    """
    
    def detect_file_fields(self, model) -> Tuple[List[str], List[str]]:
        """
        Automatically detect file fields in a Django model.
        
        Args:
            model: Django model class to analyze
            
        Returns:
            Tuple of (single_file_fields, multiple_file_fields)
        """
        single_file_fields = []
        multiple_file_fields = []
        
        for field in model._meta.get_fields():
            if isinstance(field, (models.FileField, models.ImageField)):
                single_file_fields.append(field.name)
                
            elif isinstance(field, models.ManyToManyField):
                related_model = field.related_model
                if related_model and self._has_file_fields(related_model):
                    multiple_file_fields.append(field.name)
                    
            # elif isinstance(field, models.ManyToOneRel):
            #     related_model = field.related_model
            #     if related_model and self._has_file_fields(related_model):
            #         multiple_file_fields.append(field.get_accessor_name())
                    
            # elif isinstance(field, models.OneToOneField):
            #     related_model = field.related_model
            #     if related_model and self._has_file_fields(related_model):
            #         single_file_fields.append(field.name)
                    
            # elif isinstance(field, models.OneToOneRel):
            #     related_model = field.related_model
            #     if related_model and self._has_file_fields(related_model):
            #         single_file_fields.append(field.get_accessor_name())
                
        return single_file_fields, multiple_file_fields
    
    def _has_file_fields(self, model) -> bool:
        """Check if a model has any file fields."""
        for field in model._meta.get_fields():
            if isinstance(field, (models.FileField, models.ImageField)):
                return True
        return False


# Legacy function for backward compatibility
def detect_file_fields(model) -> Tuple[List[str], List[str]]:
    """
    Legacy function for detecting file fields.
    
    This function is kept for backward compatibility.
    Use FileFieldDetector.detect_file_fields() for new code.
    """
    detector = FileFieldDetector()
    return detector.detect_file_fields(model)


from typing import Dict, List, Tuple, Any, Optional
from asgiref.sync import sync_to_async

from django.db import models

from ..file_upload import FileUploadConfig, FileFieldDetector


class BaseFileHandler:
    """Base class for file upload handlers."""
    
    def __init__(self, file_upload_config: Optional[FileUploadConfig] = None):
        self.file_upload_config = file_upload_config
        self.detector = FileFieldDetector()
    
    def _extract_files_from_request(self, request, model_name: str) -> Tuple[Dict[str, Any], Dict[str, List]]:
        """Extract single and multiple files from request."""
        single_files = {}
        multiple_files = {}
        
        if not self.file_upload_config:
            return single_files, multiple_files
        
        single_file_fields = self.file_upload_config.get_model_file_fields(model_name)
        for field_name in single_file_fields:
            if field_name in request.FILES:
                single_files[field_name] = request.FILES[field_name]
        
        multiple_file_fields = self.file_upload_config.get_model_multiple_file_fields(model_name)
        for field_name in multiple_file_fields:
            files = request.FILES.getlist(field_name)
            if files:
                multiple_files[field_name] = files
        
        return single_files, multiple_files
    
    def _get_relation_info(self, model, field_name: str) -> Optional[Dict[str, Any]]:
        """Get information about a relation field."""
        try:
            relation = next((f for f in model._meta.get_fields() if f.name == field_name), None)
            if not relation:
                return None
            
            relation_info = {
                'field': relation,
                'target_model': None,
                'relation_type': None,
                'fk_name': None
            }
            
            if isinstance(relation, models.ManyToManyField):
                relation_info['target_model'] = relation.remote_field.model
                relation_info['relation_type'] = 'many_to_many'
            elif isinstance(relation, models.ManyToOneRel):
                relation_info['target_model'] = relation.related_model
                relation_info['relation_type'] = 'many_to_one_rel'
                relation_info['fk_name'] = relation.field.name
            elif isinstance(relation, models.OneToOneField):
                relation_info['target_model'] = relation.related_model
                relation_info['relation_type'] = 'one_to_one'
            elif isinstance(relation, models.OneToOneRel):
                relation_info['target_model'] = relation.related_model
                relation_info['relation_type'] = 'one_to_one_rel'
                relation_info['fk_name'] = relation.field.name
            else:
                return None
            
            return relation_info
        except Exception:
            return None


class SyncFileHandler(BaseFileHandler):
    """Handles file upload processing for sync routes."""
    
    def process_create_files(self, request, payload, model) -> Tuple[Dict[str, Any], Dict[str, List]]:
        """Process files for create operation."""
        data = payload.model_dump()
        model_name = model.__name__
        
        single_files, multiple_files = self._extract_files_from_request(request, model_name)
        
        data.update(single_files)
        
        for field_name in multiple_files.keys():
            data.pop(field_name, None)
        
        return data, multiple_files
    
    def process_update_files(self, request, payload, model) -> Tuple[Dict[str, Any], Dict[str, List]]:
        """Process files for update operation."""
        data = payload.model_dump(exclude_unset=True)
        model_name = model.__name__
        
        single_files, multiple_files = self._extract_files_from_request(request, model_name)
        
        data.update(single_files)
        
        for field_name in multiple_files.keys():
            data.pop(field_name, None)
        
        return data, multiple_files
    
    def handle_file_relations(self, instance, file_fields_map: Dict[str, List], model) -> None:
        """Handle file relations for an instance."""
        for field_name, files in file_fields_map.items():
            relation_info = self._get_relation_info(model, field_name)
            if not relation_info:
                continue
            
            target_model = relation_info['target_model']
            relation_type = relation_info['relation_type']
            
            single_file_fields, _ = self.detector.detect_file_fields(target_model)
            if not single_file_fields:
                continue
            
            file_field = single_file_fields[0]
            
            if relation_type == 'many_to_many':
                self._handle_many_to_many_files(instance, field_name, files, target_model, file_field)
            elif relation_type == 'many_to_one_rel':
                self._handle_many_to_one_files(instance, files, target_model, file_field, relation_info['fk_name'])
            elif relation_type == 'one_to_one':
                self._handle_one_to_one_files(instance, field_name, files, target_model, file_field)
            elif relation_type == 'one_to_one_rel':
                self._handle_one_to_one_rel_files(instance, files, target_model, file_field, relation_info['fk_name'])
    
    def _handle_many_to_many_files(self, instance, field_name: str, files: List, target_model, file_field: str):
        """Handle many-to-many file relations."""
        manager = getattr(instance, field_name)
        manager.clear()
        
        created_objs = []
        for f in files:
            obj = target_model.objects.create(**{file_field: f})
            created_objs.append(obj)
        
        manager.add(*created_objs)
    
    def _handle_many_to_one_files(self, instance, files: List, target_model, file_field: str, fk_name: str):
        """Handle many-to-one file relations."""
        for f in files:
            target_model.objects.create(**{file_field: f, fk_name: instance})
    
    def _handle_one_to_one_files(self, instance, field_name: str, files: List, target_model, file_field: str):
        """Handle one-to-one file relations."""
        if files:
            related_obj = target_model.objects.create(**{file_field: files[0]})
            setattr(instance, field_name, related_obj)
            instance.save()
    
    def _handle_one_to_one_rel_files(self, instance, files: List, target_model, file_field: str, fk_name: str):
        """Handle one-to-one reverse file relations."""
        if files:
            target_model.objects.create(**{file_field: files[0], fk_name: instance})


class AsyncFileHandler(BaseFileHandler):
    """Handles file upload processing for async routes."""
    
    async def process_create_files(self, request, payload, model) -> Tuple[Dict[str, Any], Dict[str, List]]:
        """Process files for create operation asynchronously."""
        data = payload.model_dump()
        model_name = model.__name__
        
        single_files, multiple_files = self._extract_files_from_request(request, model_name)
        
        data.update(single_files)
        
        for field_name in multiple_files.keys():
            data.pop(field_name, None)
        
        return data, multiple_files
    
    async def process_update_files(self, request, payload, model) -> Tuple[Dict[str, Any], Dict[str, List]]:
        """Process files for update operation asynchronously."""
        data = payload.model_dump(exclude_unset=True)
        model_name = model.__name__
        
        single_files, multiple_files = self._extract_files_from_request(request, model_name)
        
        data.update(single_files)
        
        for field_name in multiple_files.keys():
            data.pop(field_name, None)
        
        return data, multiple_files
    
    async def handle_file_relations(self, instance, file_fields_map: Dict[str, List], model) -> None:
        """Handle file relations for an instance asynchronously."""
        get_fields = sync_to_async(lambda m: m._meta.get_fields())
        model_fields = await get_fields(model)
        
        for field_name, files in file_fields_map.items():
            relation = next((f for f in model_fields if f.name == field_name), None)
            if not relation:
                continue
            
            relation_info = await sync_to_async(self._get_relation_info)(model, field_name)
            if not relation_info:
                continue
            
            target_model = relation_info['target_model']
            relation_type = relation_info['relation_type']
            
            detect_files = sync_to_async(self.detector.detect_file_fields)
            single_file_fields, _ = await detect_files(target_model)
            if not single_file_fields:
                continue
            
            file_field = single_file_fields[0]
            
            if relation_type == 'many_to_many':
                await self._handle_many_to_many_files_async(instance, field_name, files, target_model, file_field)
            elif relation_type == 'many_to_one_rel':
                await self._handle_many_to_one_files_async(instance, files, target_model, file_field, relation_info['fk_name'])
            elif relation_type == 'one_to_one':
                await self._handle_one_to_one_files_async(instance, field_name, files, target_model, file_field)
            elif relation_type == 'one_to_one_rel':
                await self._handle_one_to_one_rel_files_async(instance, files, target_model, file_field, relation_info['fk_name'])
    
    async def _handle_many_to_many_files_async(self, instance, field_name: str, files: List, target_model, file_field: str):
        """Handle many-to-many file relations asynchronously."""
        manager = getattr(instance, field_name)
        clear_manager = sync_to_async(manager.clear)
        await clear_manager()
        
        created_objs = []
        for f in files:
            create_related = sync_to_async(lambda model, **kwargs: model.objects.create(**kwargs))
            obj = await create_related(target_model, **{file_field: f})
            created_objs.append(obj)
        
        add_to_manager = sync_to_async(lambda mgr, objs: mgr.add(*objs))
        await add_to_manager(manager, created_objs)
    
    async def _handle_many_to_one_files_async(self, instance, files: List, target_model, file_field: str, fk_name: str):
        """Handle many-to-one file relations asynchronously."""
        for f in files:
            create_related = sync_to_async(lambda model, **kwargs: model.objects.create(**kwargs))
            await create_related(target_model, **{file_field: f, fk_name: instance})
    
    async def _handle_one_to_one_files_async(self, instance, field_name: str, files: List, target_model, file_field: str):
        """Handle one-to-one file relations asynchronously."""
        if files:
            create_related = sync_to_async(lambda model, **kwargs: model.objects.create(**kwargs))
            related_obj = await create_related(target_model, **{file_field: files[0]})
            
            setattr(instance, field_name, related_obj)
            save_instance = sync_to_async(lambda obj: obj.save())
            await save_instance(instance)
    
    async def _handle_one_to_one_rel_files_async(self, instance, files: List, target_model, file_field: str, fk_name: str):
        """Handle one-to-one reverse file relations asynchronously."""
        if files:
            create_related = sync_to_async(lambda model, **kwargs: model.objects.create(**kwargs))
            await create_related(target_model, **{file_field: files[0], fk_name: instance})


from typing import List, Any, Dict, Optional
from asgiref.sync import sync_to_async

from ninja import Form
from ninja.pagination import paginate

from .base import BaseModelRouter
from ..handlers.response import AsyncResponseHandler
from ..handlers.file_handler import AsyncFileHandler
from ..utils.hooks import AsyncHookExecutor
from ..utils.model import AsyncModelUtils
from ..helpers import QuerysetFilter, parse_model_id
from ..errors import handle_exception_async


class AsyncModelRouter(BaseModelRouter):
    """
    Async implementation of model router.

    Handles all async routes registration and request processing.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.response_handler = AsyncResponseHandler()
        self.file_handler = AsyncFileHandler(self.file_upload_config)
        self.hook_executor = AsyncHookExecutor()
        self.model_utils = AsyncModelUtils()
        self.queryset_filter = QuerysetFilter(self.model)

    def register_list_route(self) -> None:
        """Register async list route with pagination and filtering."""

        @self.router.get(
            "/",
            response=List[self.list_schema],
            tags=self.get_tags(),
            operation_id=self.get_operation_id("list")
        )
        @paginate(self.paginator_class)
        async def list_items(
            request,
            q: Optional[str] = None,
            sort: Optional[str] = None,
            order: Optional[str] = "asc",
            **kwargs: Any
        ) -> List[Any]:
            """List objects with optional filtering and sorting."""
            try:
                all_items = await self.model_utils.get_all_objects(self.model)

                if self.pre_list:
                    hook_result = await self.hook_executor.execute(self.pre_list, request, all_items)
                    if hook_result is not None:
                        all_items = hook_result
                
                if q or sort or kwargs:
                    all_items = await self.queryset_filter.apply_filters_async(
                        all_items, q, sort, order, **kwargs
                    )
                else:
                    all_items = await sync_to_async(list)(all_items)

                serialized_items = []
                for item in all_items:
                    serialized = await self.model_utils.serialize_model_instance(item)
                    serialized_items.append(serialized)

                if self.custom_response:
                    return await sync_to_async(self.custom_response)(request, serialized_items)
                
                return serialized_items
            except Exception as e:
                return await handle_exception_async(e)
            
    def register_detail_route(self) -> None:
        """Register async detail route."""

        @self.router.get(
            "/{item_id}",
            response=self.detail_schema,
            tags=self.get_tags(),
            operation_id=self.get_operation_id("get")
        )
        async def get_item(request, item_id: str) -> Any:
            """Retrieve a single object by ID."""
            try:
                item_id_value = parse_model_id(self.model, item_id)
                instance = await self.model_utils.get_object_or_404(self.model, id=item_id_value)
                return await self.response_handler.handle_response(
                    instance, self.detail_schema, self.custom_response, request
                )
            except Exception as e:
                return await handle_exception_async(e)
            
    def register_create_route(self) -> None:
        """Register async create route."""
        if self.use_multipart_create:
            self._register_multipart_create_route()
        else:
            self._register_json_create_route()

    def _register_json_create_route(self) -> None:
        """Register JSON-based create route."""

        @self.router.post(
            "/",
            response=self.detail_schema,
            tags=self.get_tags(),
            operation_id=self.get_operation_id("create")
        )
        async def create_item(request, payload: self.create_schema) -> Any: # type: ignore
            """Create a new object."""
            try:
                if self.before_create:
                    payload = await self.hook_executor.execute(
                        self.before_create, request, payload, self.create_schema
                    ) or payload

                data = await self.model_utils.convert_foreign_keys(self.model, payload.model_dump())

                instance = await self.model_utils.create_instance(self.model, **data)

                if self.after_create:
                    instance = await self.hook_executor.execute(
                        self.after_create, request, instance
                    ) or instance

                return await self.response_handler.handle_response(
                    instance, self.detail_schema, self.custom_response, request
                )
            except Exception as e:
                return await handle_exception_async(e)
            
    def _register_multipart_create_route(self) -> None:
        """Register multipart/form-data create route"""

        @self.router.post(
            "/",
            response=self.detail_schema,
            tags=self.get_tags(),
            operation_id=self.get_operation_id("create")
        )
        async def create_item(request, payload: self.create_schema = Form(...)) -> Any: # type: ignore
            """Create a new object with file upload support."""
            try:
                if self.before_create:
                    payload = await self.hook_executor.execute(
                        self.before_create, request, payload, self.create_schema
                    ) or payload

                data, file_fields_map = await self.file_handler.process_create_files(
                    request, payload, self.model
                )

                data = await self.model_utils.convert_foreign_keys(self.model, data)

                instance = await self.model_utils.create_instance(self.model, **data)

                if file_fields_map:
                    await self.file_handler.handle_file_relations(
                        instance, file_fields_map, self.model
                    )

                if self.after_create:
                    instance = await self.hook_executor.execute(
                        self.after_create, request, instance
                    ) or instance

                return await self.response_handler.handle_response(
                    instance, self.detail_schema, self.custom_response, request
                )
            except Exception as e:
                return await handle_exception_async(e)
    
    def register_update_route(self) -> None:
        """Register async update route."""
        if self.use_multipart_update:
            self._register_multipart_update_route()
        else:
            self._register_json_update_route()

    def _register_json_update_route(self) -> None:
        """Register JSON-based update route."""

        @self.router.patch(
            "/{item_id}",
            response=self.detail_schema,
            tags=self.get_tags(),
            operation_id=self.get_operation_id("update")
        )
        async def update_item(request, item_id: str, payload: self.update_schema) -> Any: # type: ignore
            """Update an existing object"""
            try:
                item_id_value = parse_model_id(self.model, item_id)
                instance = await self.model_utils.get_object_or_404(self.model, id=item_id_value)

                if self.before_update:
                    payload = await self.hook_executor.execute(
                        self.before_update, request, instance, payload, self.update_schema
                    ) or payload

                data = await self.model_utils.convert_foreign_keys(
                    self.model, payload.model_dump(exclude_unset=True)
                )

                await self.model_utils.update_instance(instance, data)

                if self.after_update:
                    instance = await self.hook_executor.execute(
                        self.after_update, request, instance
                    ) or instance

                return await self.response_handler.handle_response(
                    instance, self.detail_schema, self.custom_response, request
                )
            except Exception as e:
                return await handle_exception_async(e)
            
    def _register_multipart_update_route(self) -> None:
        """Register multipart/form-data update route."""
        
        @self.router.patch(
            "/{item_id}", 
            response=self.detail_schema, 
            tags=self.get_tags(), 
            operation_id=self.get_operation_id("update")
        )
        async def update_item(request, item_id: str, payload: self.update_schema = Form(...)) -> Any: # type: ignore
            """Update an existing object with file upload support."""
            try:
                item_id_value = parse_model_id(self.model, item_id)
                instance = await self.model_utils.get_object_or_404(self.model, id=item_id_value)

                if self.before_update:
                    payload = await self.hook_executor.execute(
                        self.before_update, request, instance, payload, self.update_schema
                    ) or payload
                
                data, file_fields_map = await self.file_handler.process_update_files(
                    request, payload, self.model
                )
                
                data = await self.model_utils.convert_foreign_keys(self.model, data)
                
                await self.model_utils.update_instance(instance, data)
                
                if file_fields_map:
                    await self.file_handler.handle_file_relations(
                        instance, file_fields_map, self.model
                    )
                
                if self.after_update:
                    instance = await self.hook_executor.execute(
                        self.after_update, request, instance
                    ) or instance
                
                return await self.response_handler.handle_response(
                    instance, self.detail_schema, self.custom_response, request
                )
            except Exception as e:
                return await handle_exception_async(e)
    
    def register_delete_route(self) -> None:
        """Register async delete route."""
        
        @self.router.delete(
            "/{item_id}", 
            response={200: Dict[str, str]}, 
            tags=self.get_tags(), 
            operation_id=self.get_operation_id("delete")
        )
        async def delete_item(request, item_id: str) -> Dict[str, str]:
            """Delete an object."""
            try:
                item_id_value = parse_model_id(self.model, item_id)
                instance = await self.model_utils.get_object_or_404(self.model, id=item_id_value)

                if self.before_delete:
                    await self.hook_executor.execute(self.before_delete, request, instance)
                
                await self.model_utils.delete_instance(instance)
                
                if self.after_delete:
                    await self.hook_executor.execute(self.after_delete, instance)
                
                return {"message": f"{self.model.__name__} with ID {item_id} has been deleted"}
            except Exception as e:
                return await handle_exception_async(e)


            



"""
Schema generation utilities for lazy-ninja.
"""
from typing import Type, List, Optional, Any, Dict, cast
from pydantic import BaseModel, ConfigDict, create_model, model_validator

from django.db import models
from ninja import Schema

from .base import get_pydantic_type, serialize_model_instance


def generate_schema(
    model: Type[models.Model],
    exclude: Optional[List[str]] = None,
    optional_fields: Optional[List[str]] = None,
    update: bool = False,
) -> Type[BaseModel]:
    """Generate a Pydantic schema based on a Django model.
    
    Args:
        model: Django model class
        exclude: List of field names to exclude from schema
        optional_fields: List of field names that should be optional
        update: Whether this is an update schema (all fields optional)
        
    Returns:
        Dynamically created Pydantic schema class
        
    Note on typing:
        The return type is dynamically created and cannot be fully
        type-checked. Use type: ignore[attr-defined] when accessing
        fields on instances of the returned schema.
        
    Example:
        UserSchema = generate_schema(User, exclude=["password"])
        user_data = UserSchema(username="john")  # ✅ Works
        # user_data.username  # type: ignore[attr-defined] if needed
    """
    exclude = exclude or []
    optional_fields = optional_fields or []
    
    fields: Dict[str, Any] = {}
    for field in model._meta.fields:  # type: ignore[attr-defined]
        field_name = getattr(field, "name", None)
        if not field_name or field_name in exclude:
            continue
        
        # Get Pydantic type for this field
        pydantic_type = get_pydantic_type(field)
        
        # Check if field allows null
        field_null = getattr(field, "null", False)
        
        if update:
            # For update schemas, all fields are optional
            fields[field_name] = (Optional[pydantic_type], None)
        elif field_name in optional_fields or field_null:
            # Mark field as optional if explicitly specified or Django field allows null
            fields[field_name] = (Optional[pydantic_type], None)
        else:
            # Required field
            fields[field_name] = (pydantic_type, ...)
    
    class DynamicSchema(Schema):
        """Base schema with model serialization validator."""
        
        @model_validator(mode="before")
        def pre_serialize(cls, values: Any) -> Any:
            """Convert Django model instance to dict.
            
            Args:
                values: Input values (Django model or dict)
            
            Returns:
                Dictionary representation of the model
            """
            # Check if it's a Django model instance
            if hasattr(values, "_meta"):
                return serialize_model_instance(values)
            return values
        
        model_config = ConfigDict(from_attributes=True)

    # Create dynamic schema class
    schema = create_model(
        model.__name__ + "Schema",
        __base__=DynamicSchema,
        **fields  # type: ignore[arg-type]
    )
    
    return cast(Type[BaseModel], schema)



"""JWT token helpers and stateful blacklist logic."""

from datetime import datetime, timedelta, timezone
from typing import Any, Dict
from uuid import uuid4

import jwt
from django.core.cache import cache
from jwt import ExpiredSignatureError, PyJWTError
from ninja.errors import HttpError

from .config import (
    get_jwt_secret,
    get_jwt_algorithm,
    get_jwt_issuer,
    get_jwt_audience,
    get_blacklist_prefix,
    is_stateful,
)


def generate_token(user: Any, *, expires_in: int, token_type: str) -> str:
    """Generate JWT token for user."""
    now = datetime.now(timezone.utc)
    payload = {
        "sub": str(user.id),  # type: ignore[attr-defined]
        "type": token_type,
        "iat": int(now.timestamp()),
        "exp": int((now + timedelta(seconds=expires_in)).timestamp()),
        "iss": get_jwt_issuer(),
        "aud": get_jwt_audience(),
        "jti": uuid4().hex,
    }
    return jwt.encode(payload, get_jwt_secret(), algorithm=get_jwt_algorithm())


def decode_raw_token(token: str) -> Dict[str, Any]:
    try:
        payload = jwt.decode(
            token,
            get_jwt_secret(),
            algorithms=[get_jwt_algorithm()],
            issuer=get_jwt_issuer(),
            audience=get_jwt_audience(),
        )
    except ExpiredSignatureError as exc:
        raise HttpError(401, "Expired token.") from exc
    except PyJWTError as exc:
        raise HttpError(401, "Invalid token.") from exc

    return payload


def validate_token_payload(payload: Dict[str, Any], expected_type: str) -> None:
    if payload.get("type") != expected_type:
        raise HttpError(401, "Invalid token type.")

    if is_stateful():
        jti = payload.get("jti")
        if not jti:
            raise HttpError(401, "Missing token identifier.")
        if is_token_blacklisted(str(jti)):
            raise HttpError(401, "Token revoked.")


def decode_token(token: str, expected_type: str) -> Dict[str, Any]:
    payload = decode_raw_token(token)
    validate_token_payload(payload, expected_type)
    return payload


def get_token_ttl(payload: Dict[str, Any]) -> int:
    exp = payload.get("exp")
    try:
        ttl = int(exp) - int(datetime.now(timezone.utc).timestamp()) # type: ignore
        return max(ttl, 0)
    except (TypeError, ValueError):
        return 0


def blacklist_key(jti: str) -> str:
    return f"{get_blacklist_prefix()}:{get_jwt_issuer()}:{get_jwt_audience()}:{jti}"


def is_token_blacklisted(jti: str) -> bool:
    return bool(cache.get(blacklist_key(jti)))


def blacklist_token_payload(payload: Dict[str, Any]) -> None:
    if not is_stateful():
        return
    jti = payload.get("jti")
    if not jti:
        return
    ttl = get_token_ttl(payload)
    if ttl <= 0:
        return
    cache.set(blacklist_key(str(jti)), True, timeout=ttl)



from typing import Dict, Any, Optional
import traceback
import logging
from django.conf import settings

from django.db import DatabaseError
from django.core.exceptions import (
    ValidationError as DjangoValidationError,
    ObjectDoesNotExist,
    SynchronousOnlyOperation,
)
from django.http import JsonResponse
from ninja.errors import HttpError
from colorama import Fore, Style, init

init(autoreset=True) 
logger = logging.getLogger(__name__)

class LazyNinjaError(Exception):
    """Base exception class for Lazy Ninja errors."""
    status_code = 500
    default_message = "An unexpected error occurred"

    def __init__(self, message: Optional[str] = None, status_code: Optional[int] = None):
        self.message = message or self.default_message
        if status_code is not None:
            self.status_code = status_code
        super().__init__(self.message)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "error": {
                "status_code": self.status_code,
                "message": self.message,
                "type": self.__class__.__name__,
            }
        }

class SynchronousOperationError(LazyNinjaError):
    status_code = 500
    default_message = "Synchronous operation called from async context - use sync_to_async"

class DatabaseOperationError(LazyNinjaError):
    status_code = 500
    default_message = "Database operation failed"

class ValidationError(LazyNinjaError):
    status_code = 400
    default_message = "Validation error"

class NotFoundError(LazyNinjaError):
    status_code = 404
    default_message = "Resource not found"

class PermissionDeniedError(LazyNinjaError):
    status_code = 403
    default_message = "Permission denied"

class BadRequestError(LazyNinjaError):
    status_code = 400
    default_message = "Bad request"

class ConflictError(LazyNinjaError):
    status_code = 409
    default_message = "Resource conflict"

def handle_exception(exc: Exception) -> JsonResponse:
    if isinstance(exc, ObjectDoesNotExist) or "matches the given query" in str(exc):
        error = NotFoundError(str(exc))

    elif isinstance(exc, PermissionError) or "permission" in str(exc).lower():
        error = PermissionDeniedError(str(exc))

    elif isinstance(exc, DjangoValidationError) or isinstance(exc, ValueError):
        error = ValidationError(str(exc))

    elif isinstance(exc, DatabaseError):
        error = DatabaseOperationError(str(exc))

    elif isinstance(exc, SynchronousOnlyOperation):
        error = SynchronousOperationError()

    elif isinstance(exc, HttpError):
        error = LazyNinjaError(str(exc), exc.status_code)

    elif isinstance(exc, LazyNinjaError):
        error = exc

    else:
        error = LazyNinjaError(str(exc))

    print(
        f"{Fore.RED + Style.BRIGHT}[LazyNinja ❌] {error.__class__.__name__}: "
        f"{Fore.YELLOW}{error.message}"
    )
    
    if getattr(settings, "DEBUG", False):
        print(Fore.LIGHTBLACK_EX + traceback.format_exc())

    return JsonResponse(error.to_dict(), status=error.status_code)


async def handle_exception_async(exc: Exception) -> JsonResponse:
    return handle_exception(exc)


from typing import Any, Callable, Optional
from asgiref.sync import sync_to_async


class BaseHookExecutor:
    """Base class for hook execution."""

    def _is_valid_hook(self, hook: Optional[Callable]) -> bool:
        """Check if hook is valid and not a default hook."""
        return hook is not None and not getattr(hook, "__is_default_hook__", False)
    

class SyncHookExecutor(BaseHookExecutor):
    """Handles hook execution for sync routes."""

    def execute(self, hook: Optional[Callable], *args, **kwargs) -> Any:
        """
        Safely execute a hook function if it exists and is not a default hook.

        Args:
            hook: The hook function to execute
            *args: Positional arguments to pass to the hook
            **kwargs: Keyword arguments to pass to the hook

        Returns:
            The result of the hook execution or None
        """
        if self._is_valid_hook(hook):
            return hook(*args, **kwargs)
        return None
    

class AsyncHookExecutor(BaseHookExecutor):
    """Handles hook execution for async routes."""

    async def execute(self, hook: Optional[Callable], *args, **kwargs) -> Any:
        """
        Safely execute a hook function asynchronously.
        
        Args:
            hook: The hook function to execute
            *args: Positional arguments to pass to the hook
            **kwargs: Keyword arguments to pass to the hook
            
        Returns:
            The result of the hook execution or None
        """
        if self._is_valid_hook(hook):
            return await sync_to_async(hook)(*args, **kwargs)
        return None
    

# Legacy function wrappers for backward compatibility
def execute_hook(hook: Optional[Callable], *args, **kwargs) -> Any:
    """Legacy wrapper for SyncHookExecutor.execute"""
    executor = SyncHookExecutor()
    return executor.execute(hook, *args, **kwargs)


async def execute_hook_async(hook: Optional[Callable], *args, **kwargs) -> Any:
    """Legacy wrapper for AsyncHookExecutor.execute"""
    executor = AsyncHookExecutor()
    return await executor.execute(hook, *args, **kwargs)


def get_hook(controller: Optional[object], hook_name: str, passed_hook: Optional[Callable] = None) -> Optional[Callable]:
    if controller:
        controller_hook = getattr(controller, hook_name, None)
        if controller_hook and not getattr(controller_hook, "__is_default_hook__", False):
                return controller_hook
    return passed_hook


"""Auth route registration and handlers."""

import logging
from typing import Any, Dict, Optional, cast

from django.contrib.auth import authenticate, get_user_model
from django.contrib.auth.password_validation import validate_password
from django.core.exceptions import ValidationError
from django.db import IntegrityError, transaction
from django.http import HttpResponse
from django.utils import timezone as dj_timezone
from ninja import NinjaAPI, Schema
from ninja.errors import HttpError

from .config import (
    cookie_secure_flag,
    get_login_fields,
    get_token_lifetimes,
    should_log_auth_events,
    should_rotate_refresh,
    should_set_cookies,
    should_validate_password,
    is_stateful,
)
from .hooks import (
    on_login_hook,
    on_register_hook,
    on_refresh_hook,
    on_logout_hook,
)
from .tokens import (
    generate_token,
    decode_token,
    blacklist_token_payload,
)
from ..utils.schema import generate_schema
from ..utils.type_guards import (
    has_user_field,
    get_user_identifier,
)

logger = logging.getLogger(__name__)


def register_auth_routes(
    api: NinjaAPI,
    *,
    access_cookie_name: str = "lazy_ninja_access_token",
    refresh_cookie_name: str = "lazy_ninja_refresh_token",
    cookie_path: str = "/",
    tags: Optional[list[str]] = None,
) -> None:
    """
    Registers a full JWT-based authentication flow on the given NinjaAPI.

    Endpoints added:
        POST /auth/login
        POST /auth/register
        POST /auth/refresh
        GET  /auth/me
        POST /auth/logout
    """
    auth_tags = tags or ["Auth"]

    User = get_user_model()
    lifetimes = get_token_lifetimes()

    def _has_username_field() -> bool:
        return has_user_field(User, "username")

    UserPublicSchema = generate_schema(User, exclude=["password"])

    class TokenPairSchema(Schema):
        access: str
        refresh: str

    class AuthResponseSchema(TokenPairSchema):
        user: dict

    class LoginSchema(Schema):
        login: Optional[str] = None
        email: Optional[str] = None
        username: Optional[str] = None
        password: str

    class RegisterSchema(Schema):
        login: Optional[str] = None
        email: Optional[str] = None
        password: str
        username: Optional[str] = None
        first_name: Optional[str] = None
        last_name: Optional[str] = None

    class RefreshSchema(Schema):
        refresh: Optional[str] = None

    class MeResponseSchema(Schema):
        user: dict

    def _serialize_user(user: Any) -> dict:
        """Serialize user to dict using dynamic schema."""
        schema_instance = UserPublicSchema.model_validate(user)
        return cast(dict, schema_instance.model_dump())

    def _apply_auth_cookies(response: HttpResponse, access: str, refresh: Optional[str]) -> None:
        if not should_set_cookies():
            return
        secure_flag = cookie_secure_flag()
        response.set_cookie(
            access_cookie_name,
            access,
            max_age=lifetimes["access"],
            httponly=True,
            secure=secure_flag,
            samesite="Lax",
            path=cookie_path,
        )
        if refresh:
            response.set_cookie(
                refresh_cookie_name,
                refresh,
                max_age=lifetimes["refresh"],
                httponly=True,
                secure=secure_flag,
                samesite="Lax",
            )

    def _clear_auth_cookies(response: HttpResponse) -> None:
        if not should_set_cookies():
            return
        response.delete_cookie(access_cookie_name, path=cookie_path)
        response.delete_cookie(refresh_cookie_name, path=cookie_path)

    def _build_auth_payload(user: Any, *, rotate_refresh: bool = True) -> dict:
        access = generate_token(user, expires_in=lifetimes["access"], token_type="access")
        refresh = (
            generate_token(user, expires_in=lifetimes["refresh"], token_type="refresh")
            if rotate_refresh else None
        )
        return {
            "access": access,
            "refresh": refresh or "",
            "user": _serialize_user(user),
        }

    def _get_username_field_name() -> str:
        return str(getattr(User, "USERNAME_FIELD", "username"))

    def _resolve_login_identifier(payload: LoginSchema) -> str:
        login_fields = get_login_fields()
        if "login" in login_fields and payload.login:
            return payload.login
        if "email" in login_fields and payload.email:
            return payload.email
        if "username" in login_fields and payload.username:
            return payload.username
        raise HttpError(400, "Missing login identifier.")

    def _resolve_register_identifier(payload: RegisterSchema, username_field: str) -> str:
        identifier = payload.username or payload.login
        if username_field == "email":
            identifier = payload.email or payload.login
        if not identifier:
            raise HttpError(400, "Missing registration identifier.")
        return identifier

    def _authenticate_user(request, identifier: str, password: str):
        login_fields = get_login_fields()
        username_field = _get_username_field_name()
        user = authenticate(request, **{username_field: identifier}, password=password)
        if user or not has_user_field(User, "email"):
            return user
        if "email" in login_fields and username_field != "email" and "@" in identifier:
            user_by_email = User.objects.filter(email__iexact=identifier).first()
            if user_by_email is not None:
                username_value = user_by_email.get_username()
                return authenticate(request, username=username_value, password=password)
            return authenticate(request, email=identifier, password=password)
        return user

    def _token_from_request(request, expected_type: str) -> Optional[str]:
        cookie_val = (
            request.COOKIES.get(access_cookie_name)
            if expected_type == "access"
            else request.COOKIES.get(refresh_cookie_name)
        )

        auth_header = request.headers.get("Authorization") or ""
        if auth_header.startswith("Bearer "):
            return auth_header.split(" ", 1)[1]

        return cookie_val

    def _get_user_from_authorization(request, expected_type: str = "access") -> Any:
        token = _token_from_request(request, expected_type)
        if not token:
            raise HttpError(401, "Missing credentials")

        payload = decode_token(token, expected_type)
        user_id = payload.get("sub")

        try:
            user = User.objects.get(id=user_id)  # type: ignore[misc]
        except User.DoesNotExist as exc:  # type: ignore[attr-defined]
            raise HttpError(401, "User not found.") from exc
        return user

    def _build_response(request, payload: dict, *, status: int = 200) -> HttpResponse:
        response = api.create_response(request, payload, status=status)
        _apply_auth_cookies(response, payload["access"], payload.get("refresh"))
        return response

    @api.post("/auth/login", response=AuthResponseSchema, tags=auth_tags)
    def login(request, payload: LoginSchema):
        identifier = _resolve_login_identifier(payload)
        user = _authenticate_user(request, identifier, payload.password)
        if not user:
            if should_log_auth_events():
                logger.warning(
                    "Failed login attempt for email: %s from IP: %s",
                    identifier,
                    request.META.get("REMOTE_ADDR", "unknown")
                )
            raise HttpError(401, "Invalid credentials.")

        user.last_login = dj_timezone.now()  # type: ignore[attr-defined]
        user.save(update_fields=["last_login"])

        if should_log_auth_events():
            user_identifier = (
                getattr(user, "email", None) or
                getattr(user, "username", "unknown")
            )
            logger.info(
                "User logged in: %s (ID: %s) from IP: %s",
                user_identifier,
                user.id,  # type: ignore [attr-defined]
                request.META.get("REMOTE_ADDR", "unknown")
            )

        on_login = on_login_hook()
        if on_login:
            on_login(user=user, request=request)

        return _build_response(request, _build_auth_payload(user))

    @api.post("/auth/register", response=AuthResponseSchema, tags=auth_tags)
    def register(request, payload: RegisterSchema):
        username_field = _get_username_field_name()
        identifier = _resolve_register_identifier(payload, username_field)

        normalized_email = ""
        if has_user_field(User, "email"):
            email_value = payload.email or (identifier if username_field == "email" else None)
            if not email_value:
                raise HttpError(400, "E-mail is required.")
            normalized_email = email_value.lower().strip()
            if User.objects.filter(email__iexact=normalized_email).exists():
                raise HttpError(400, "E-mail already registered.")

        username = payload.username or (normalized_email or identifier)

        if _has_username_field() and User.objects.filter(username__iexact=username).exists():  # type: ignore
            raise HttpError(400, "Username already taken.")

        if should_validate_password():
            try:
                validate_password(payload.password)
            except ValidationError as exc:
                error_msg = "; ".join(exc.messages) if exc.messages else "Invalid password"
                raise HttpError(400, f"Password validation failed: {error_msg}")

        try:
            with transaction.atomic():
                user_kwargs: Dict[str, Any] = {
                    "password": payload.password,
                }

                if has_user_field(User, "email"):
                    user_kwargs["email"] = normalized_email
                if _has_username_field():
                    user_kwargs["username"] = username
                if has_user_field(User, username_field) and username_field not in user_kwargs:
                    user_kwargs[username_field] = identifier

                if payload.first_name is not None and hasattr(User, "first_name"):
                    user_kwargs["first_name"] = payload.first_name
                if payload.last_name is not None and hasattr(User, "last_name"):
                    user_kwargs["last_name"] = payload.last_name

                user = User.objects.create_user(**user_kwargs)  # type: ignore[misc]

                if should_log_auth_events():
                    logger.info(
                        "New user registered: %s from IP: %s",
                        get_user_identifier(user),
                        request.META.get("REMOTE_ADDR", "unknown")
                    )

                on_register = on_register_hook()
                if on_register:
                    on_register(user=user, request=request)

        except IntegrityError as exc:
            logger.error("Registration integrity error: %s", str(exc))
            raise HttpError(400, "Registration failed. Email or username may already be in use.")

        return _build_response(request, _build_auth_payload(user))

    @api.post("/auth/refresh", response=TokenPairSchema, tags=auth_tags)
    def refresh_token(request, payload: RefreshSchema):
        token_source = payload.refresh or _token_from_request(request, expected_type="refresh")
        if not token_source:
            raise HttpError(401, "Renewal token missing.")

        data = decode_token(token_source, expected_type="refresh")
        user_id = data.get("sub")

        try:
            user = User.objects.only("id").get(id=user_id)  # type: ignore[misc]
        except User.DoesNotExist as exc:  # type: ignore[attr-defined]
            raise HttpError(401, "User not found.") from exc

        if is_stateful():
            blacklist_token_payload(data)

        token_pair = {
            "access": generate_token(user, expires_in=lifetimes["access"], token_type="access"),
            "refresh": generate_token(
                user,
                expires_in=lifetimes["refresh"],
                token_type="refresh",
            ) if should_rotate_refresh() else "",
        }

        if should_log_auth_events():
            logger.debug("Token refreshed for user ID: %s", user.id)  # type: ignore[attr-defined]

        response = api.create_response(request, token_pair, status=200)
        _apply_auth_cookies(response, token_pair["access"], token_pair["refresh"])

        on_refresh = on_refresh_hook()
        if on_refresh:
            on_refresh(user=user, request=request)
        return response

    @api.get("/auth/me", response=MeResponseSchema, tags=auth_tags)
    def me(request):
        user = _get_user_from_authorization(request, expected_type="access")
        return {"user": _serialize_user(user)}

    @api.post("/auth/logout", tags=auth_tags)
    def logout(request):
        try:
            if should_log_auth_events():
                token = _token_from_request(request, "access")
                if token:
                    payload = decode_token(token, "access")
                    user_id = payload.get("sub")
                    logger.info(
                        "User logged out: ID %s from IP: %s",
                        user_id,
                        request.META.get("REMOTE_ADDR", "unknown")
                    )
                    if is_stateful():
                        blacklist_token_payload(payload)

            if is_stateful():
                refresh_token_value = _token_from_request(request, "refresh")
                if refresh_token_value:
                    refresh_payload = decode_token(refresh_token_value, "refresh")
                    blacklist_token_payload(refresh_payload)
        except Exception:
            pass

        response = api.create_response(request, {"detail": "Logged out"}, status=200)
        _clear_auth_cookies(response)

        on_logout = on_logout_hook()
        if on_logout:
            on_logout(request=request)
        return response


import json

import pytest
from django.contrib.auth import get_user_model
from django.core.cache import cache
from django.test.utils import override_settings


@pytest.mark.django_db
def test_register_success(client):
    payload = {
        "email": "user1@example.com",
        "password": "S0mePassw0rd!",
        "username": "user1",
        "first_name": "User",
        "last_name": "One",
    }

    response = client.post(
        "/api/auth/register",
        data=json.dumps(payload),
        content_type="application/json",
    )

    assert response.status_code == 200
    data = response.json()
    assert "access" in data
    assert "refresh" in data
    assert "user" in data
    assert data["user"]["email"] == payload["email"]


@pytest.mark.django_db
def test_register_duplicate_email_fails(client):
    User = get_user_model()
    User.objects.create_user(
        username="userdup",
        email="dup@example.com",
        password="S0mePassw0rd!",
    )

    response = client.post(
        "/api/auth/register",
        data=json.dumps({
            "email": "dup@example.com",
            "password": "S0mePassw0rd!",
        }),
        content_type="application/json",
    )

    assert response.status_code == 400


@pytest.mark.django_db
def test_login_success_and_me_flow(client):
    with override_settings(LAZY_NINJA_AUTH={"LOGIN_FIELDS": ["email", "username", "login"]}):
        User = get_user_model()
        User.objects.create_user(
            username="user2@example.com",
            email="user2@example.com",
            password="S0mePassw0rd!",
        )
        login_response = client.post(
            "/api/auth/login",
            data=json.dumps({
                "email": "user2@example.com",
                "password": "S0mePassw0rd!",
            }),
            content_type="application/json",
        )

        assert login_response.status_code == 200
        login_data = login_response.json()
        access_token = login_data["access"]
        assert access_token

        me_response = client.get(
            "/api/auth/me",
            HTTP_AUTHORIZATION=f"Bearer {access_token}",
        )

        assert me_response.status_code == 200
        me_data = me_response.json()
        assert me_data["user"]["email"] == "user2@example.com"


@pytest.mark.django_db
def test_login_invalid_credentials(client):
    with override_settings(LAZY_NINJA_AUTH={"LOGIN_FIELDS": ["email", "username", "login"]}):
        response = client.post(
            "/api/auth/login",
            data=json.dumps({
                "email": "unknown@example.com",
                "password": "badpass",
            }),
            content_type="application/json",
        )

        assert response.status_code == 401


@pytest.mark.django_db
def test_refresh_token_flow(client):
    with override_settings(LAZY_NINJA_AUTH={"LOGIN_FIELDS": ["email", "username", "login"]}):
        User = get_user_model()
        User.objects.create_user(
            username="user3@example.com",
            email="user3@example.com",
            password="S0mePassw0rd!",
        )

        login_response = client.post(
            "/api/auth/login",
            data=json.dumps({
                "email": "user3@example.com",
                "password": "S0mePassw0rd!",
            }),
            content_type="application/json",
        )

        assert login_response.status_code == 200
        refresh_token = login_response.json()["refresh"]

        refresh_response = client.post(
            "/api/auth/refresh",
            data=json.dumps({"refresh": refresh_token}),
            content_type="application/json",
        )

        assert refresh_response.status_code == 200
        refresh_data = refresh_response.json()
        assert "access" in refresh_data
        assert "refresh" in refresh_data


@pytest.mark.django_db
def test_logout_clears_cookies(client):
    with override_settings(LAZY_NINJA_AUTH={"LOGIN_FIELDS": ["email", "username", "login"]}):
        User = get_user_model()
        User.objects.create_user(
            username="user4@example.com",
            email="user4@example.com",
            password="S0mePassw0rd!",
        )

        login_response = client.post(
            "/api/auth/login",
            data=json.dumps({
                "email": "user4@example.com",
                "password": "S0mePassw0rd!",
            }),
            content_type="application/json",
        )

        assert login_response.status_code == 200
        access_cookie = login_response.cookies.get("lazy_ninja_access_token")
        refresh_cookie = login_response.cookies.get("lazy_ninja_refresh_token")
        assert access_cookie is not None
        assert refresh_cookie is not None

        logout_response = client.post(
            "/api/auth/logout",
            data=json.dumps({}),
            content_type="application/json",
        )

        assert logout_response.status_code == 200
        assert logout_response.cookies.get("lazy_ninja_access_token") is not None
        assert logout_response.cookies.get("lazy_ninja_refresh_token") is not None


@pytest.mark.django_db
def test_login_with_username_identifier(client):
    with override_settings(LAZY_NINJA_AUTH={"LOGIN_FIELDS": ["username"]}):
        User = get_user_model()
        User.objects.create_user(
            username="userlogin",
            email="userlogin@example.com",
            password="S0mePassw0rd!",
        )

        response = client.post(
            "/api/auth/login",
            data=json.dumps({
                "username": "userlogin",
                "password": "S0mePassw0rd!",
            }),
            content_type="application/json",
        )

        assert response.status_code == 200
        assert response.json()["access"]


@pytest.mark.django_db
def test_login_with_login_identifier(client):
    with override_settings(LAZY_NINJA_AUTH={"LOGIN_FIELDS": ["login", "email"]}):
        User = get_user_model()
        User.objects.create_user(
            username="loginalias",
            email="loginalias@example.com",
            password="S0mePassw0rd!",
        )

        response = client.post(
            "/api/auth/login",
            data=json.dumps({
                "login": "loginalias@example.com",
                "password": "S0mePassw0rd!",
            }),
            content_type="application/json",
        )

        assert response.status_code == 200
        assert response.json()["access"]


@pytest.mark.django_db
def test_issuer_audience_mismatch_rejected(client):
    with override_settings(LAZY_NINJA_AUTH={"LOGIN_FIELDS": ["email"]}):
        User = get_user_model()
        User.objects.create_user(
            username="issuser",
            email="issuser@example.com",
            password="S0mePassw0rd!",
        )

        login_response = client.post(
            "/api/auth/login",
            data=json.dumps({
                "email": "issuser@example.com",
                "password": "S0mePassw0rd!",
            }),
            content_type="application/json",
        )

        assert login_response.status_code == 200
        access_token = login_response.json()["access"]

    with override_settings(LAZY_NINJA_AUTH={"JWT_ISS": "other-iss", "JWT_AUD": "other-aud"}):
        me_response = client.get(
            "/api/auth/me",
            HTTP_AUTHORIZATION=f"Bearer {access_token}",
        )

    assert me_response.status_code == 401


@pytest.mark.django_db
def test_stateful_refresh_blacklists_refresh_token(client):
    cache.clear()
    User = get_user_model()
    User.objects.create_user(
        username="stateful1",
        email="stateful1@example.com",
        password="S0mePassw0rd!",
    )

    with override_settings(LAZY_NINJA_AUTH={"STATEFUL": True, "LOGIN_FIELDS": ["email"]}):
        login_response = client.post(
            "/api/auth/login",
            data=json.dumps({
                "email": "stateful1@example.com",
                "password": "S0mePassw0rd!",
            }),
            content_type="application/json",
        )

        assert login_response.status_code == 200
        refresh_token = login_response.json()["refresh"]

        first_refresh = client.post(
            "/api/auth/refresh",
            data=json.dumps({"refresh": refresh_token}),
            content_type="application/json",
        )
        assert first_refresh.status_code == 200

        second_refresh = client.post(
            "/api/auth/refresh",
            data=json.dumps({"refresh": refresh_token}),
            content_type="application/json",
        )
        assert second_refresh.status_code == 401


@pytest.mark.django_db
def test_stateful_logout_blacklists_access_token(client):
    cache.clear()
    User = get_user_model()
    User.objects.create_user(
        username="stateful2",
        email="stateful2@example.com",
        password="S0mePassw0rd!",
    )

    with override_settings(LAZY_NINJA_AUTH={"STATEFUL": True, "LOGIN_FIELDS": ["email"]}):
        login_response = client.post(
            "/api/auth/login",
            data=json.dumps({
                "email": "stateful2@example.com",
                "password": "S0mePassw0rd!",
            }),
            content_type="application/json",
        )
        assert login_response.status_code == 200
        access_token = login_response.json()["access"]

        logout_response = client.post(
            "/api/auth/logout",
            data=json.dumps({}),
            content_type="application/json",
            HTTP_AUTHORIZATION=f"Bearer {access_token}",
        )
        assert logout_response.status_code == 200

        me_response = client.get(
            "/api/auth/me",
            HTTP_AUTHORIZATION=f"Bearer {access_token}",
        )
        assert me_response.status_code == 401


from __future__ import annotations
import subprocess
from pathlib import Path
import textwrap
import shutil
import sys


def _find_project_package(project_root: Path, project_name: str) -> Path:
    """
    Find the Django project package directory based on the project name.
    This searches for the settings.py file in common project structures.
    project_root: directory that contains the created project package (where package folder lives)
    project_name: name of the Django project (used to locate settings.py)
    Returns the directory containing settings.py.
    """
    project_root = project_root.resolve()
    
    p = project_root / project_name / project_name / "settings.py"
    if p.exists():
        return p.parent

    p = project_root / project_name / "settings.py"
    if p.exists():
        return p.parent

    p = project_root / "settings.py"
    if p.exists():
        return project_root

    matches = list(project_root.rglob("settings.py"))
    if not matches:
        raise RuntimeError(
            f"Could not locate Django project package under {project_root}. "
            "Expected to find a directory containing settings.py (e.g. project_name/settings.py)."
        )

    for m in matches:
        if m.parent.name == project_name:
            return m.parent

    return matches[0].parent


def _run_django_startproject(name: str, target: str | None = None):
    """
    Run `django-admin startproject <name> [<target>]`.
    Falls back to `python -m django startproject` if `django-admin` is not in PATH.
    """
    if shutil.which("django-admin"):
        cmd = ["django-admin", "startproject", name]
        if target:
            cmd.append(target)
    else:
        print("[lazy-ninja] 'django-admin' not found on PATH, using 'python -m django startproject' fallback.")
        cmd = [sys.executable, "-m", "django", "startproject", name]
        if target:
            cmd.append(target)

    subprocess.run(cmd, check=True)


def _update_urls_py(urls_file: Path, project_name: str) -> bool:
    """
    Automatically add the API route to urls.py

    Returns True if successful, False if manual intervention is needed
    """
    if not urls_file.exists():
        print(f"[lazy-ninja]⚠️ urls.py not found at {urls_file}")
        return False
    
    try:
        content = urls_file.read_text(encoding="utf-8")

        api_import = f"from {project_name}.api import api as lazy_api"
        api_path = "path('api/', lazy_api.urls),"

        if api_import in content and api_path in content:
            print("[lazy-ninja] ✅ API routes already configured in urls.py")
            return True
        
        lines = content.split('\n')
        new_lines = []

        needs_include_import = 'include' not in content
        needs_ninja_import = api_import not in content
        needs_api_path = api_path not in content

        in_urlpatterns = False
        urlpatterns_closed = False
        added_imports = False
        added_path = False

        for i, line in enumerate(lines):
            if not added_imports and line.strip().startswith('from django'):
                new_lines.append(line)

                next_lines = lines[i+1:i+3]
                is_last_django_import = True
                for next_line in next_lines:
                    if next_line.strip().startswith('from django.'):
                        is_last_django_import = False
                        break
                
                if is_last_django_import:
                    if needs_include_import and 'from django.urls import' in line and 'include' not in line:
                        if 'path' in line:
                            new_lines[-1] = line.replace('path', 'path, include')

                    if needs_ninja_import:
                        new_lines.append(f"from {project_name}.api import api as lazy_api")

                        added_imports = True
                continue

            if 'urlpatterns = [' in line:
                in_urlpatterns = True
                new_lines.append(line)

                if needs_api_path:
                    new_lines.append(f"    path('api/', lazy_api.urls),")
                    added_path = True
                continue

            if in_urlpatterns and line.strip() == ']':
                urlpatterns_closed = True
                
                if needs_api_path and not added_path:
                    new_lines.append(f"    path('api/', lazy_api.urs),")
                    added_path = True
                new_lines.append(line)
                in_urlpatterns = False
                continue

            new_lines.append(line)

        if not added_imports:
            import_lines = []
            if needs_include_import:
                import_lines.append("from django.urls import path, include")
            if needs_ninja_import:
                import_lines.append(f"from {project_name}.api import api as lazy_api")
            
            insert_pos = 0
            for i, line in enumerate(new_lines):
                if line.strip().startswith(('import ', 'from ')):
                    insert_pos = i + 1

            for j, import_line in enumerate(import_lines):
                new_lines.insert(insert_pos + j, import_line)

        urls_file.write_text('\n'.join(new_lines), encoding="utf-8")
        print(f"[lazy-ninja] ✅ Updated {urls_file} with API routes")
        return True
        
    except Exception as e:
        print(f"[lazy-ninja] ⚠️ Could not automatically update urls.py: {e}")
        return False




def _scaffold_lazy_ninja(project_root: Path, project_name: str, title: str | None = None):
    """
    Create api.py inside the project package and append instructions to settings.py.

    project_root: directory that contains the created project package (where package folder lives)
    """
    pkg_dir = _find_project_package(project_root, project_name) 
    
    if not pkg_dir.exists():
        raise RuntimeError(f"Project package not found at expected location: {pkg_dir}")

    api_py = pkg_dir / "api.py"
    
    if api_py.exists():
        print(f"[lazy-ninja] ⚠️ api.py already exists at {api_py}, skipping creation.")
    else:
        api_content = textwrap.dedent(f"""\
            from ninja import NinjaAPI
            from lazy_ninja.builder import DynamicAPI

            # API title: {title or project_name}
            api = NinjaAPI(title="{title or project_name}")

            auto_api = DynamicAPI(
                api,
                is_async=True,
            )

            auto_api.init()
        """)
        api_py.write_text(api_content, encoding="utf-8")
        print(f"[lazy-ninja] ✅ Wrote {api_py}")
    
    urls_file = pkg_dir / "urls.py"
    urls_updated = _update_urls_py(urls_file, project_name)

    settings_file = pkg_dir / "settings.py"
    
    if settings_file.exists():
        with settings_file.open("a", encoding="utf-8") as f:
            f.write("\n\n# === Lazy Ninja scaffold ===\n")
            f.write("# A sample Lazy Ninja API has been created in api.py\n")

            if urls_updated:
                f.write("# API routes have been automatically configured at /api/\n")
                f.write("# Visit http://localhost:8000/api/docs after running 'python manage.py runserver'\n")
            else:
                f.write("# To expose it, add something like the following to your project's urls.py:\n")
                f.write("#\n")
                f.write(f"#   from django.urls import path, include\n")
                f.write(f"#   from {project_name}.api import api as lazy_api\n")
                f.write("#\n")
                f.write("#   urlpatterns = [\n")
                f.write("#       path('api/', lazy_api.urls),\n")
                f.write("#       # ... other patterns\n")
                f.write("#   ]\n")
                f.write("#\n# Install lazy-ninja (pip install lazy-ninja) and follow the README for more examples.\n")


def startproject_command(name: str, directory: str | None = None, title: str | None = None):
    """
    Create a Django project and always add a minimal Lazy Ninja scaffold (api.py + settings note).
    """
    try:
        _run_django_startproject(name, directory)
    except subprocess.CalledProcessError as e:
        print(f"[lazy-ninja] ❌ django startproject failed: {e}")
        sys.exit(e.returncode)

    project_root = Path(directory).resolve() if directory else Path(".").resolve()

    try:
        _scaffold_lazy_ninja(project_root, name, title)
        print(f"[lazy-ninja] ✅ Project '{name}' created with Lazy Ninja scaffold at {project_root / name}")
    except Exception as e:
        print(f"[lazy-ninja] ⚠️ Project created but failed to add Lazy Ninja scaffold: {e}")
        print("You can manually add api.py as shown in the README.")



import asyncio
import inflect
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Set, Dict, List, Type, Union, Any

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import connection
from django.apps import apps

from ninja import NinjaAPI
from pydantic import BaseModel

from .core import register_model_routes
from .utils import generate_schema
from .helpers import to_kebab_case
from .pagination import get_pagination_strategy
from .file_upload import FileUploadConfig, detect_file_fields
from .auth import register_auth_routes
from .utils.type_guards import get_model_field_names, has_field

p = inflect.engine()

class ExclusionConfig:
    """
    Flexible configuration for excluding models from routes registration.
    
    Supports multiple ways of specifying exclusions with a unified approach.
    """
    def __init__(
        self,
        exclude: Optional[Dict[str, Union[bool, Set[str], Any]]] = None
    ):
        """
        Initialize exclusion configuration.
        
        Args:
            exclude: configuration for model/app exclusions:
                {
                    'app_name': True, # Exclude all models from this app
                    'app_name': {'ModelName1', 'ModelName2'}, # Exclude specific models from this app
                    'app_name': None, # No exclusions for this app
                }
        By default, the system apps are excluded: auth, contenttypes, admin, sessions
        """
        self.excluded = {
            'auth': True,
            'contenttypes': True,
            'admin': True,
            'sessions': True
        }
        
        if exclude:
            self.excluded.update(exclude) # type: ignore
            
    def should_exclude_model(self, model) -> bool:
        """
        Determine if a model should be excluded from route registration.
        
        Args:
            model: Django model to check for exclusion.
        
        Returns:
            bool: True if the model should be excluded, False otherwise.
        """
        app_label = model._meta.app_label
        model_name = model.__name__
        
        app_config = self.excluded.get(app_label)
        
        if app_config is True:
            return True
        
        if app_config is None:
            return False
        
        if isinstance(app_config, set):
            return model_name in app_config
        
        return False

class DynamicAPI:
    """
    Dynamically registers CRUD routes for Django models using Django Ninja.

    This class scans installed Django models (excluding those from specified apps)
    and automatically creates/uses Pydantic schemas for listing, detailing,
    creating, and updating models. It registers these routes with Ninja.
    """

    def __init__(
        self,
        api: NinjaAPI,
        exclude: Optional[Dict[str, Union[bool, Set[str], Any]]] = None,
        schema_config: Optional[Dict[str, Dict[str, List[str]]]] = None,
        custom_schemas: Optional[Dict[str, Dict[str, Type[BaseModel]]]] = None,
        pagination_type: Optional[str] = None,
        file_fields: Optional[Dict[str, List[str]]] = None,
        auto_detect_files: bool = True,
        auto_multipart: bool = True,
        use_multipart: Optional[Dict[str, Dict[str, bool]]] = None,
        is_async: bool = True,
        auth: Optional[bool] = None,
        auth_profile_model: Optional[Type[Any]] = None,
        auth_access_cookie_name: str = "lazy_ninja_access_token",
        auth_refresh_cookie_name: str = "lazy_ninja_refresh_token",
        auth_cookie_path: str = "/",
        auth_tags: Optional[List[str]] = None,
    ):
        """
        Initializes the DynamicAPI instance.

        Args:
            api: The NinjaAPI instance.
            exclude: Configuration for model/app exclusions.
            schema_config: Dictionary mapping model names to schema configurations
                           (e.g., exclude fields and optional fields).
            custom_schemas: Dictionary mapping model names to custom Pydantic Schema classes for
                            list, detail, create, and update operations.  The dictionary should have the structure:
                            `{"ModelName": {"list": ListSchema, "detail": DetailSchema, "create": CreateSchema, "update": UpdateSchema}}`
                            If a schema is not provided for a specific operation, the default generated schema will be used.
            pagination_type: Type of pagination to use ('limit-offset' or 'page-number').
                           If None, uses NINJA_PAGINATION_CLASS from settings.
            file_fields: Dictionary mapping model names to lists of file field names
                         (e.g., {"Product": ["image", "document"]}).
            auto_detect_files: Whether to automatically detect file fields in models.
            auto_multipart: Whether to automatically use multipart for models with file fields.
            use_multipart: Dictionary specifying which models should use multipart/form-data
                           (e.g., {"Product": {"create": True, "update": True}}).
            is_async: Whether to use async routes (default: True).
            auth: Enable built-in auth routes when True. If None, falls back
                  to settings.LAZY_NINJA.get("auth", False).
            auth_profile_model: Optional Django model class for a user profile
                  (OneToOne with the user). If not provided, can be set via
                  settings.LAZY_NINJA["profile_model"] as "app_label.ModelName".
            auth_access_cookie_name: Name of the access token cookie.
            auth_refresh_cookie_name: Name of the refresh token cookie.
            auth_cookie_path: Cookie path for auth cookies.
            auth_tags: Optional list of tags for auth endpoints.
               
        Pagination Configuration:
            The pagination can be configured in three ways (in order of precedence):
            1. pagination_type parameter in DynamicAPI
            2. NINJA_PAGINATION_CLASS in Django settings
            3. Default to LimitOffsetPagination (if no settings or parameter are provided)
            
            The page size is configured via NINJA_PAGINATION_PER_PAGE in settings.
        """
        self.api = api
        self.exclusion_config = ExclusionConfig(exclude=exclude)
        self.schema_config = schema_config or {}
        self.custom_schemas = custom_schemas or {}
        self.is_async = is_async
        self.pagination_strategy = get_pagination_strategy(pagination_type=pagination_type)
       
        self.file_fields = file_fields or {}
        self.use_multipart = use_multipart or {}
        self.auto_detect_files = auto_detect_files
        self.auto_multipart = auto_multipart
        self.file_upload_config = FileUploadConfig(
            file_fields=self.file_fields
        )

        lazy_cfg = getattr(settings, "LAZY_NINJA", {}) or {}
        if auth is None:
            self.auth_enabled = bool(lazy_cfg.get("auth", False))
        else:
            self.auth_enabled = bool(auth)

        profile_model_setting = lazy_cfg.get("profile_model")
        resolved_profile_model: Optional[Type[Any]] = auth_profile_model
        if resolved_profile_model is None and profile_model_setting:
            if isinstance(profile_model_setting, str):
                try:
                    resolved_profile_model = apps.get_model(profile_model_setting)
                except LookupError:
                    raise RuntimeError(
                        f"Invalid LAZY_NINJA['profile_model']: {profile_model_setting!r}"
                    )
            elif isinstance(profile_model_setting, type):
                resolved_profile_model = profile_model_setting

        self.auth_profile_model = resolved_profile_model
        self.auth_access_cookie_name = auth_access_cookie_name
        self.auth_refresh_cookie_name = auth_refresh_cookie_name
        self.auth_cookie_path = auth_cookie_path
        self.auth_tags = auth_tags

        self._already_registered = False
        
    @staticmethod
    def _get_existing_tables() -> List[str]:
        """Get list of existing database tables.
        
        Returns:
            List of table names in the database.
        """
        with connection.cursor() as cursor:
            return connection.introspection.table_names(cursor)
    
    def _register_all_models_sync(self) -> None:
        existing_tables = self._get_existing_tables()
        try:
            user_model = get_user_model()
        except Exception:  # pragma: nocover - defensive
            user_model = None

        for model in apps.get_models():
            # Safe access to model meta
            app_label = model._meta.app_label  # type: ignore[attr-defined]
            model_name = model.__name__

            if not getattr(self, "auth_enabled", False) and user_model and model is user_model:
                continue

            if self.exclusion_config.should_exclude_model(model):
                continue
            
            # Safe access to db_table
            db_table = model._meta.db_table  # type: ignore[attr-defined]
            if db_table not in existing_tables:
                continue

            custom_schema = self.custom_schemas.get(model_name)

            if custom_schema:
                list_schema = custom_schema.get("list") or generate_schema(model)
                detail_schema = custom_schema.get("detail") or generate_schema(model) 
                create_schema = custom_schema.get("create") 
                update_schema = custom_schema.get("update") 

            else:
                model_config = self.schema_config.get(model_name, {})
                default_excludes = [
                    "id",
                    "created_at",
                    "updated_at",
                    "deleted_at",
                ]
                raw_excludes = model_config.get("exclude", default_excludes)
                exclude_fields = [field for field in raw_excludes if has_field(model, field)]
                
                optional_fields = model_config.get("optional_fields", [])

                list_schema = generate_schema(model)
                detail_schema = generate_schema(model)
                create_schema = generate_schema(model, exclude=exclude_fields, optional_fields=optional_fields)
                update_schema = generate_schema(model, exclude=exclude_fields, optional_fields=optional_fields, update=True)

            detected_single_file_fields = []
            detected_multiple_file_fields = []
            
            if self.auto_detect_files:
                detected_single_file_fields, detected_multiple_file_fields = detect_file_fields(model)
                
            provided_file_fields = list(self.file_fields.get(model_name, []) or [])
            existing_field_names = get_model_field_names(model, exclude_relations=True)
            sanitized_provided = [field for field in provided_file_fields if field in existing_field_names]
            model_file_fields = list(set(sanitized_provided + detected_single_file_fields))
            
            if model_file_fields:
                self.file_upload_config.file_fields[model_name] = model_file_fields
                
            if detected_multiple_file_fields:
                existing = self.file_upload_config.multiple_file_fields.get(model_name, [])
                self.file_upload_config.multiple_file_fields[model_name] = list(set(existing + detected_multiple_file_fields))
            
            use_multipart_create = self.use_multipart.get(model_name, {}).get("create", False)
            use_multipart_update = self.use_multipart.get(model_name, {}).get("update", False)
        
            has_any_file_fields = bool(model_file_fields) or bool(detected_multiple_file_fields)
            if self.auto_multipart and has_any_file_fields:
                use_multipart_create = True
                use_multipart_update = True
                
            register_model_routes(
                api=self.api,
                model=model,
                base_url=f"/{p.plural(to_kebab_case(model_name))}", # type: ignore
                list_schema=list_schema,
                detail_schema=detail_schema, 
                create_schema=create_schema,
                update_schema=update_schema,
                pagination_strategy=self.pagination_strategy, # type: ignore
                file_upload_config=self.file_upload_config if model_file_fields else None,
                use_multipart_create=use_multipart_create,
                use_multipart_update=use_multipart_update,
                is_async=getattr(self, 'is_async', True)
            )
            
    def register_all_models(self) -> None:
        """
        Scans Django models and registers routes.

        Excludes models from specified apps.  Uses custom schemas if provided;
        otherwise, generates schemas based on schema_config or defaults.
        """
        if self._already_registered:
            return
        
        self._already_registered = True
        try:
            loop = asyncio.get_running_loop()
            with ThreadPoolExecutor() as executor:
                future = executor.submit(self._register_all_models_sync)
                future.result()
                
        except RuntimeError:
            self._register_all_models_sync()
            
    def init(self) -> None:
        """
        Initializes the DynamicAPI.
        
        This method acts as the public initializer for the library. It internally calls
        register_all_models() to scan models and register their corresponding CRUD routes.
        """
        if getattr(self, "auth_enabled", False):
            auth_kwargs: Dict[str, Any] = {}
            if self.auth_profile_model is not None:
                auth_kwargs["profile_model"] = self.auth_profile_model
            if self.auth_access_cookie_name:
                auth_kwargs["access_cookie_name"] = self.auth_access_cookie_name
            if self.auth_refresh_cookie_name:
                auth_kwargs["refresh_cookie_name"] = self.auth_refresh_cookie_name
            if self.auth_cookie_path:
                auth_kwargs["cookie_path"] = self.auth_cookie_path
            if self.auth_tags is not None:
                auth_kwargs["tags"] = self.auth_tags

            register_auth_routes(self.api, **auth_kwargs)

        self.register_all_models()



from typing import Any, Optional, Type

from django.db.models import Model
from ninja import NinjaAPI
from pydantic import BaseModel

from .base import BaseModelController
from .helpers import get_hook
from .registry import ModelRegistry, controller_for
from .routes import register_model_routes_internal
from .file_upload import FileUploadConfig

def register_model_routes(
    api: NinjaAPI,
    model: Type[Model],
    base_url: str,
    list_schema: Type[BaseModel],
    detail_schema: Type[BaseModel],
    create_schema: Optional[Type[BaseModel]] = None,
    update_schema: Optional[Type[BaseModel]] = None,
    pagination_strategy: Optional[Any] = None,
    file_upload_config: Optional[FileUploadConfig] = None,
    use_multipart_create: bool = False,
    use_multipart_update: bool = False,
    is_async: bool = True,
) -> None:
    """Register CRUD routes for a Django model using Django Ninja.

    This function retrieves the registered controller for the model (if any)
    and passes its hooks to the internal route registration function.

    Args:
        api: NinjaAPI instance
        model: Django model class
        base_url: Base URL for the routes (e.g., "/users")
        list_schema: Pydantic schema for list responses
        detail_schema: Pydantic schema for detail responses
        create_schema: Optional Pydantic schema for create requests
        update_schema: Optional Pydantic schema for update requests
        pagination_strategy: Optional pagination strategy instance
        file_upload_config: Optional configuration for file uploads
        use_multipart_create: Whether to use multipart/form-data for create
        use_multipart_update: Whether to use multipart/form-data for update
        is_async: Whether to use async routes (default: True)
    
    Example:
        >>> from myapp.models import User
        >>> UserSchema = generate_schema(User)
        >>> register_model_routes(
        ...     api=api,
        ...     model=User,
        ...     base_url="/users",
        ...     list_schema=UserSchema,
        ...     detail_schema=UserSchema,
        ... )
    """
    ModelRegistry.discover_controllers()

    controller = ModelRegistry.get_controller(model.__name__)
    if not controller:
        controller = BaseModelController
    
    register_model_routes_internal(
        api=api,
        model=model,
        base_url=base_url,
        list_schema=list_schema,
        detail_schema=detail_schema,
        create_schema=create_schema,
        update_schema=update_schema,
        pre_list=get_hook(controller, 'pre_list'),
        before_create=get_hook(controller, 'before_create'),
        after_create=get_hook(controller, 'after_create'),
        before_update=get_hook(controller, 'before_update'),
        after_update=get_hook(controller, 'after_update'),
        before_delete=get_hook(controller, 'before_delete'),
        after_delete=get_hook(controller, 'after_delete'),
        custom_response=get_hook(controller, 'custom_response'),
        pagination_strategy=pagination_strategy,
        file_upload_config=file_upload_config,
        use_multipart_create=use_multipart_create,
        use_multipart_update=use_multipart_update,
        is_async=is_async
    )


from types import SimpleNamespace

from ninja import Schema

from lazy_ninja.core import register_model_routes
from lazy_ninja.base import BaseModelController
from lazy_ninja.registry import ModelRegistry


class ListSchema(Schema):
    id: int


class DetailSchema(Schema):
    id: int


class CreateSchema(Schema):
    id: int


class UpdateSchema(Schema):
    id: int | None = None


def _call_register_model_routes(monkeypatch, controller):
    captured = {}

    def fake_register(**kwargs):
        captured.update(kwargs)

    monkeypatch.setattr(ModelRegistry, "discover_controllers", lambda: None)
    monkeypatch.setattr(ModelRegistry, "get_controller", lambda name: controller)

    from lazy_ninja import core as core_module

    monkeypatch.setattr(core_module, "register_model_routes_internal", fake_register)

    register_model_routes(
        api=SimpleNamespace(),
        model=SimpleNamespace(__name__="Widget", _meta=None),
        base_url="/widgets",
        list_schema=ListSchema,
        detail_schema=DetailSchema,
        create_schema=CreateSchema,
        update_schema=UpdateSchema,
        pagination_strategy=None,
    )

    return captured


def test_register_model_routes_uses_custom_controller_hooks(monkeypatch):
    class CustomController(BaseModelController):
        @classmethod
        def before_create(cls, request, payload, schema):
            return payload

    captured = _call_register_model_routes(monkeypatch, CustomController)
    assert captured["before_create"].__func__ is CustomController.before_create.__func__
    assert captured["after_create"].__func__ is BaseModelController.after_create.__func__


def test_register_model_routes_fallback_to_default_hooks(monkeypatch):
    captured = _call_register_model_routes(monkeypatch, None)
    assert captured["before_create"].__func__ is BaseModelController.before_create.__func__
    assert captured["after_delete"].__func__ is BaseModelController.after_delete.__func__



import pytest
from ninja.conf import settings as ninja_settings
from ninja.pagination import LimitOffsetPagination, PageNumberPagination

from lazy_ninja.pagination import (
    get_pagination_strategy,
    LimitOffsetPaginationStrategy,
    PageNumberPaginationStrategy,
)


def test_get_pagination_strategy_defaults_to_limit_offset(monkeypatch):
    monkeypatch.setattr(ninja_settings, "PAGINATION_CLASS", None, raising=False)
    strategy = get_pagination_strategy()
    assert isinstance(strategy, LimitOffsetPaginationStrategy)
    assert strategy.get_paginator() is LimitOffsetPagination


def test_get_pagination_strategy_respects_settings(monkeypatch):
    monkeypatch.setattr(
        ninja_settings,
        "PAGINATION_CLASS",
        "ninja.pagination.PageNumberPagination",
        raising=False,
    )
    strategy = get_pagination_strategy()
    assert isinstance(strategy, PageNumberPaginationStrategy)
    assert strategy.get_paginator() is PageNumberPagination


def test_get_pagination_strategy_explicit_type():
    strategy = get_pagination_strategy(pagination_type="page-number")
    assert isinstance(strategy, PageNumberPaginationStrategy)
    assert strategy.get_pagination_class_name() == "PageNumberPagination"


def test_get_pagination_strategy_invalid_type():
    with pytest.raises(ValueError):
        get_pagination_strategy(pagination_type="unknown")



from __future__ import annotations

import contextlib
import io

import pytest

from lazy_ninja.cli import main as main_mod
from lazy_ninja.cli import client_generator, startproject


def test_build_parser_accepts_valid_generate_client_args():
    parser = main_mod.build_parser()
    language = next(iter(client_generator.GENERATOR_CONFIG.keys()))
    args = parser.parse_args(["generate-client", language, "--settings", "myproj.settings"])
    assert args.cmd == "generate-client"
    assert args.language == language
    assert args.settings == "myproj.settings"


def test_build_parser_rejects_invalid_language():
    parser = main_mod.build_parser()
    with pytest.raises(SystemExit) as exc:
        parser.parse_args(["generate-client", "not-a-language", "--settings", "s"])
    assert exc.value.code == 2


def test_main_generate_client_requires_schema_or_settings_and_errors(monkeypatch):
    buf = io.StringIO()
    with contextlib.redirect_stderr(buf):
        with pytest.raises(SystemExit) as exc:
            main_mod.main(["generate-client", "python"]) 
    assert exc.value.code == 2 or exc.value.code == 1
    assert "either --schema or --settings must be provided" in buf.getvalue()


def test_main_generate_client_calls_handle_generate_client(monkeypatch):
    called = {}

    def fake_handle(args):
        called["args"] = args

    monkeypatch.setattr(client_generator, "handle_generate_client", fake_handle)

    language = next(iter(client_generator.GENERATOR_CONFIG.keys()))
    main_mod.main(["generate-client", language, "--settings", "myproj.settings", "--output", "dest"])

    assert "args" in called
    assert called["args"].settings == "myproj.settings"
    assert called["args"].language == language
    assert called["args"].output == "dest"


def test_main_startproject_invokes_startproject_command(monkeypatch):
    called = {}

    def fake_startproject(name, directory, title):
        called["called_with"] = (name, directory, title)

    monkeypatch.setattr(startproject, "startproject_command", fake_startproject)
    main_mod.main(["init", "myproj", "target_dir", "--title", "API Title"])

    assert "called_with" in called
    assert called["called_with"] == ("myproj", "target_dir", "API Title")


def test_main_prints_help_when_no_args(capfd):
    main_mod.main([])
    out, err = capfd.readouterr()
    assert "Lazy Ninja CLI" in out or "Generate client code" in out



from __future__ import annotations
import argparse
from pathlib import Path
from . import client_generator, startproject

def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="lazy-ninja",
        description="🌀 Lazy Ninja CLI - tools for generating clients and projects.",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    sub = parser.add_subparsers(dest="cmd")
    
    gen = sub.add_parser("generate-client", help="Generate client code from OpenAPI schema")
    gen.add_argument("language", choices=list(client_generator.GENERATOR_CONFIG.keys()))
    gen.add_argument("--output", default="./client", help="Output dir or file")
    gen.add_argument("--settings", help="Django settings module (e.g. myproject.settings). Required unless --schema is provided.")
    gen.add_argument("--api-module", default="settings.api", help="Module path where `api = NinjaAPI()` is defined")
    gen.add_argument("--api-var", default="api", help="Name of the Ninja API variable in module (default: api)")
    gen.add_argument("--schema", type=Path, help="Path to pre-generated OpenAPI JSON (skip Django setup)")
    
    proj = sub.add_parser("init", help="Create a Django project scaffold preconfigured for Lazy Ninja")
    proj.add_argument("name", help="Project name")
    proj.add_argument("directory", nargs="?", default=None, help="Optional target directory (same semantics as django-admin)")
    proj.add_argument("--title", nargs="?", default=None, help="Optional API title")
    
    return parser

def main(argv: list[str] | None = None):
    parser = build_parser()
    args = parser.parse_args(argv)
    
    if args.cmd == "generate-client":
        if args.schema is None and not args.settings:
            parser.error("either --schema or --settings must be provided for generate-client")
        client_generator.handle_generate_client(args)
    elif args.cmd == "init":
        startproject.startproject_command(args.name, args.directory, args.title)
    else:
        parser.print_help()  


import os
import sys
import subprocess
import argparse
import importlib
from pathlib import Path

GENERATOR_CONFIG = {
    "javascript": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "javascript",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "typescript-types": {
        "cmd": ["npx", "openapi-typescript", "{schema}", "--output", "{out}"],
        "ext": "ts",
    },
    "dart": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "dart-dio",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "python": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "python",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "typescript-axios": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "typescript-axios",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "typescript-fetch": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "typescript-fetch",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "java": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "java",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "kotlin": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "kotlin",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "go": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "go",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "csharp": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "csharp",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "ruby": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "ruby",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "swift5": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "swift5",
            "-o", "{out_dir}"
        ],
        "ext": None
    }
}


def setup_django(settings_module: str):
    sys.path.insert(0, os.getcwd())
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", settings_module)
    
    import django
    django.setup()
    

def dump_openapi(api_module: str, api_var: str, out_file: Path):
    from django.test import RequestFactory
    from ninja.openapi.views import openapi_json
    
    mod = importlib.import_module(api_module)
    api = getattr(mod, api_var)
    
    req = RequestFactory().get("/api/openapi.json")
    resp = openapi_json(req, api)
    out_file.write_bytes(resp.content)
    

def generate_client(schema_path: Path, language: str, output: str):
    cfg = GENERATOR_CONFIG.get(language)

    cmd = [p.format(schema=str(schema_path), out=output, out_dir=output) for p in cfg["cmd"]]
    print(f"[LazyNinja] ▶ Running: {' '.join(cmd)}")
    proc = subprocess.run(cmd)
    if proc.returncode != 0:
        print(f"[LazyNinja] ❌ Generator failed.")
        sys.exit(proc.returncode)
    print(f"[LazyNinja] ✅ Client ({language}) generated at {output}")
    
def handle_generate_client(args):
    """
    args: namespace from argparse with attributes:
      - language, output, settings, api_module, schema (optional), api_var
    """
    schema_file = Path(args.schema) if getattr(args, "schema", None) else Path(".lazy_ninja_openapi.json")

    if getattr(args, "schema", None):
        if not schema_file.exists():
            print(f"[LazyNinja] ❌ Schema file not found: {schema_file}")
            sys.exit(1)
    else:
        try:
            setup_django(args.settings)
        except Exception as e:
            print("[LazyNinja] ❌ Failed to setup Django. Make sure your settings are importable.")
            print("Error:", e)
            sys.exit(1)

        try:
            dump_openapi(args.api_module, args.api_var, schema_file)
        except ModuleNotFoundError as e:
            print("[LazyNinja] ❌ Failed to import module while dumping OpenAPI schema.")
            print("Missing module:", e.name)
            print("Consider generating a schema file with your project dependencies installed, then pass --schema path/to/schema.json")
            sys.exit(1)
        except Exception as e:
            print("[LazyNinja] ❌ Failed to generate schema.")
            print("Error:", e)
            sys.exit(1)

    try:
        generate_client(schema_file, args.language, args.output)
    finally:
        if not getattr(args, "schema", None) and schema_file.exists():
            schema_file.unlink(missing_ok=True)


"""
Core utility functions for lazy-ninja.
These are the fundamental building blocks used throughout the library.
"""
import asyncio
from typing import Type, Any, Dict
from decimal import Decimal
from asgiref.sync import sync_to_async

from django.db import models
from django.shortcuts import get_object_or_404



def convert_foreign_keys(model: Type[models.Model], data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Converts integer values for ForeignKey fields in `data` to the corresponding model instances.
    
    Args:
        model: Django model class
        data: Dictionary with field data
        
    Returns:
        Dictionary with ForeignKey integers converted to model instances
    """
    for field in model._meta.fields:
        if isinstance(field, models.ForeignKey) and field.name in data:
            fk_value = data[field.name]
            if isinstance(fk_value, (int, str)):
                data[field.name] = field.related_model.objects.get(pk=fk_value) # type: ignore
    return data


def get_field_value_safely(obj: models.Model, field: models.Field) -> Any:
    """
    Get field value without triggering database queries.
    For ForeignKey fields, returns the ID directly.
    
    Args:
        obj: Django model instance
        field: Django model field
        
    Returns:
        Field value or None if not accessible
    """
    if isinstance(field, models.ForeignKey):
        attname = field.attname
        if hasattr(obj, attname):
            return getattr(obj, attname)
        return None
    
    # For non-ForeignKey fields, get the value directly
    try:
        value = getattr(obj, field.name)
        return value
    except Exception:
        return None


def serialize_model_instance(obj: models.Model) -> Dict[str, Any]:
    """
    Serializes a Django model instance into a dictionary with simple types.
    Avoids triggering database queries for related fields.
    
    Args:
        obj: Django model instance
        
    Returns:
        Dictionary with serialized field values
    """
    data = {}
    for field in obj._meta.fields:
        try:
            value = get_field_value_safely(obj, field)
            
            if value is None:
                data[field.name] = None
            elif isinstance(field, models.UUIDField):
                data[field.name] = str(value)
            elif isinstance(field, (models.DateField, models.DateTimeField)):
                data[field.name] = value.isoformat() if value else None
            elif isinstance(field, (models.ImageField, models.FileField)):
                data[field.name] = value.url if hasattr(value, 'url') else str(value)
            elif isinstance(field, models.ForeignKey):
                target_field = field.target_field
                if isinstance(target_field, models.UUIDField):
                    data[field.name] = str(value) if value else None
                else:
                    data[field.name] = value
            elif hasattr(value, 'pk'):
                data[field.name] = value.pk
            else:
                data[field.name] = value
        except Exception:
            data[field.name] = None
    return data


def is_async_context() -> bool:
    """
    Check if we're in an async context by inspecting the call stack.
    
    Returns:
        True if in async context, False otherwise
    """
    try:
        return asyncio.get_event_loop().is_running()
    except RuntimeError:
        return False


def get_pydantic_type(field: models.Field) -> Type:
    """
    Map a Django model field to an equivalent Python type for Pydantic validation.
    
    Args:
        field: Django model field
        
    Returns:
        Python type for Pydantic
    """
    if isinstance(field, models.UUIDField):
        return str
    elif isinstance(field, models.AutoField):
        return int
    elif isinstance(field, (models.CharField, models.TextField)):
        return str
    elif isinstance(field, models.IntegerField):
        return int
    elif isinstance(field, models.DecimalField):
        return Decimal
    elif isinstance(field, models.FloatField):
        return float
    elif isinstance(field, models.BooleanField):
        return bool
    elif isinstance(field, (models.DateField, models.DateTimeField)):
        return str
    elif isinstance(field, (models.ImageField, models.FileField)):
        return str
    elif isinstance(field, models.ForeignKey):
        target_field = field.target_field
        return get_pydantic_type(target_field)
    else:
        return str


# Async versions of core functions
convert_foreign_keys_async = sync_to_async(convert_foreign_keys)
serialize_model_instance_async = sync_to_async(serialize_model_instance)
get_all_objects_async = sync_to_async(lambda m: m.objects.all())
get_object_or_404_async = sync_to_async(get_object_or_404)


